"""
Objects to perform ASE calculations with machine-learned potentials.
"""
from __future__ import annotations

import sys
import os
import pickle
import json
import itertools
import warnings
import dataclasses
import numpy as np
import pandas as pd
#import abipy.tools.duck as duck

from pathlib import Path
from multiprocessing import Pool
#from typing import Type, Any, Optional, Union
from monty.json import MontyEncoder
from monty.functools import lazy_property
from monty.collections import dict2namedtuple
from pymatgen.core.lattice import Lattice
from pymatgen.util.coord import pbc_shortest_vectors

from abipy.core import Structure
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt #, get_axarray_fig_plt,
from abipy.tools.iotools import workdir_with_prefix, PythonScript, ShellScript
from abipy.tools.printing import print_dataframe
from abipy.ml.aseml import (relax_atoms, get_atoms, as_calculator, ase_optimizer_cls, RX_MODE, fix_atoms,
                            MlNeb, MlGsList, CalcBuilder, make_ase_neb, nprocs_for_ntasks)


@dataclasses.dataclass
class Entry:
    """
    Stores the relaxed structure with the associated energy and the Cartesian forces.
    """
    isite: int             # Index of the site being relaxed.
    structure: Structure   # pymatgen Structure
    energy: float          # Energy in eV
    forces: np.ndarray     # Forces in eV/Ang

    @classmethod
    def from_atoms_and_calculator(cls, isite, atoms, calculator) -> Entry:
        """
        Build an Entry from an Atoms object by attaching a calculator
        to compute energy and forces at fixed geometry.
        """
        structure = Structure.as_structure(atoms)
        # NB: Keep sites within the unit cell so that we can compare entries in __eq__
        structure.translate_sites(range(len(structure)), np.zeros(3), to_unit_cell=True)
        return cls(isite=isite,
                   structure=structure,
                   energy=float(calculator.get_potential_energy(atoms=atoms)),
                   forces=calculator.get_forces(atoms=atoms),
                   )

    @classmethod
    def from_traj(cls, isite, traj) -> Entry:
        """
        Build an Entry by taking the last Atoms object from an ASE trajectory.
        """
        atoms = traj[-1]
        structure = Structure.as_structure(atoms)
        # NB: Keep sites within the unit cell so that we can compare entries in __eq__
        structure.translate_sites(range(len(structure)), np.zeros(3), to_unit_cell=True)
        return cls(isite=isite,
                   structure=structure,
                   energy=float(atoms.get_potential_energy()),
                   forces=atoms.get_forces(),
                   )

    def __eq__(self, other: Entry) -> bool:
        """
        Invoked by python to evaluate `self == other`.
        """
        return abs(self.energy - other.energy) / len(self.structure) < 1e-4 and \
               np.abs(self.structure.lattice.matrix - other.structure.lattice.matrix).max() < 1e-3 and \
               all(np.abs(site1.frac_coords - site2.frac_coords).max() < 1e-3
                   for site1, site2 in zip(self.structure, other.structure))


@dataclasses.dataclass
class Pair:
    """
    Stores info on a possible transition between two relaxed configurations.
    """
    index1: int                # Index of first configuration in entries
    index2: int
    ediff: float               # Energy difference in eV
    dist:  float               # Distance between sites in Ang.
    frac_coords1: np.ndarray   # Fractional coords of sites.
    frac_coords2: np.ndarray

    def get_dict4pandas(self) -> dict:
        """Useful to construct pandas DataFrames"""
        return dataclasses.asdict(self)


class RelaxScanner:
    """
    This object employs an ASE calculator to perform many
    structural relaxations in which the initial configurations are obtained
    by displacing or inserting an atom on a grid covering the unit cell.
    The relaxed configurations are then compared with each other and only
    the unique solutions (Entry objects) are kept and stored to disk in pickle format.
    """

    @classmethod
    def pickle_load(cls, workdir) -> RelaxScanner:
        """
        Reconstruct the object from a pickle file located in workdir.
        """
        with open(Path(workdir) / f"{cls.__name__}.pickle", "rb") as fh:
            return pickle.load(fh)

    def __init__(self, structure, isite, mesh, nn_name,
                 relax_mode: str = "ions", fmax: float = 1e-3, steps=500,
                 verbose: int = 0, optimizer_name="BFGS", pressure=0.0,
                 workdir=None, prefix=None):
        """
        Args:
            structure: Structure object or any file supported by pymatgen providing a structure.
            isite: Index of the site to be displaced or string with chemical element to be added.
            mesh: Size of the mesh along the smallest lattice vector.
            nn_name: String with the name of the NN potential e.g. "chgnet" or "m3gnet".
            relax_mode: "ions" to relax ions only, "cell" for ions + cell, "no" for no relaxation.
            fmax: total force tolerance for relaxation convergence.
                Here fmax is a sum of force and stress forces. Defaults to 0.1.
            steps: max number of steps for relaxation.
            verbose: Verbosity level.
            optimizer_name: name of the ASE optimizer class to use
            pressure: Target pressure.
            workdir: Working directory.
        """
        self.initial_structure = Structure.as_structure(structure).copy()

        if str(isite).isdigit():
            self.isite = int(isite)
        else:
            # Assume isite is a chemical element to be added at the origin.
            self.initial_structure.insert(0, isite, [0, 0, 0])
            self.isite = 0

        assert len(self.initial_structure) > self.isite >= 0

        # Compute 3d mesh
        abc = np.array(self.initial_structure.lattice.abc)
        delta = abc.min() / mesh
        self.nx, self.ny, self.nz = [int(_) for _ in np.rint(abc / delta)]

        self.nn_name = nn_name
        self.relax_mode = relax_mode
        RX_MODE.validate(self.relax_mode)
        self.fmax = fmax
        self.steps = steps
        self.verbose = verbose
        self.optimizer_name = optimizer_name
        self.pressure = pressure

        self.workdir = workdir_with_prefix(workdir, prefix)

        # Write pickle file for object persistence.
        with open(self.workdir / f"{self.__class__.__name__}.pickle", "wb") as fh:
            pickle.dump(self, fh)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation with verbosity level `verbose`."""
        return f"""\

{self.__class__.__name__} parameters:

     site index  = {self.isite}
     nx          = {self.nx}
     ny          = {self.ny}
     nz          = {self.nz}
     relax_mode  = {self.relax_mode}
     fmax        = {self.fmax}
     steps       = {self.steps}
     optimizer   = {self.optimizer_name}
     pressure    = {self.pressure}
     calculator  = {self.nn_name}
     verbose     = {self.verbose}
     workdir     = {str(self.workdir)}

=== INITIAL STRUCTURE ===

{self.initial_structure}
"""

    def _make_directory(self, start, stop) -> Path:
        """
        Create sub-directory inside the workdir to perform relaxations
        for a subset of structures defined by `start` and `stop`.
        Return path to directory.
        """
        # Make sure we are not already running similar ranges.
        top = str(self.workdir)
        dirpaths = [name for name in os.listdir(top) if os.path.isdir(os.path.join(top, name))
                    and name.startswith("start_")]
        for dpath in dirpaths:
            # Directory name has pattern: f"start:{start}-stop:{stop}"
            # Directory name has pattern: f"start_{start}-stop_{stop}"
            tokens = dpath.split("-")
            this_start = int(tokens[0].replace("start_", ""))
            this_stop = int(tokens[1].replace("stop_", ""))
            if start >= this_start and start < this_stop:
                raise RuntimeError(f"{start=}, {stop=} range overlaps with {dpath=}")

        # Make sure we are not already running the same range.
        newdir = self.workdir / f"start_{start}-stop_{stop}"
        if newdir.exists():
            raise RuntimeError(f"{newdir=} already exists!")

        newdir.mkdir()
        return newdir

    def get_atoms_with_frac_coords(self, frac_coords, with_fixex_atoms=True) -> Atoms:
        """
        Return Atoms instance with frac_coords at site index `isite`.
        By default, Fixed Contraints are applied to all atoms except isite.
        """
        structure = self.initial_structure.copy()
        structure._sites[self.isite].frac_coords = np.array(frac_coords)
        atoms = get_atoms(structure)
        if with_fixex_atoms:
            fix_atoms(atoms, fix_inds=[i for i in range(len(atoms)) if i != self.isite])
        return atoms

    def run_start_count(self, start: int, count: int) -> Path:
        """
        Invoke ASE to relax `count` structures starting from index `start`.
        Stores unique results in an internal list and save results to file in pickle format.
        Return Path to the directory with the results.

        Args:
            start: Index of initial structure.
            count: Number of structures to relax. If < 0, all structures are considered.
        """
        if count < 0:
            count = self.nx * self.ny * self.nz

        stop = start + count
        directory = self._make_directory(start, stop)

        #with warnings.catch_warnings():
        #    warnings.simplefilter("ignore")
        calculator = as_calculator(self.nn_name)

        entries, errors = [], []
        for cnt, (ix, iy, iz) in enumerate(itertools.product(range(self.nx), range(self.ny), range(self.nz))):
            if not (stop > cnt >= start): continue

            atoms = self.get_atoms_with_frac_coords((ix/self.nx, iy/self.ny, iz/self.nz))

            if self.relax_mode == RX_MODE.no:
                # Just GS energy, no relaxation.
                entry = Entry.from_atoms_and_calculator(self.isite, atoms, calculator)

            else:
                try:
                    # Relax atoms with constraints.
                    relax = relax_atoms(atoms,
                                        relax_mode=self.relax_mode,
                                        optimizer=self.optimizer_name,
                                        fmax=self.fmax,
                                        pressure=self.pressure,
                                        verbose=self.verbose,
                                        steps=self.steps,
                                        traj_path=directory / "relax.traj",
                                        calculator=calculator,
                    )

                    entry = Entry.from_traj(self.isite, relax.traj)

                except Exception:
                    entry = None
                    errors.append((ix, iy, iz))

            # This is the part in which duplicated entries are discarded.
            if entry is not None and entry not in entries:
                entries.append(entry)

        if errors:
            print(f"ASE relaxation raised {len(errors)} exception(s)!")

        # Save results in pickle format.
        with open(directory / "entries.pickle", "wb") as fh:
            pickle.dump(entries, fh)

        with open(directory / "COMPLETED", "wt") as fh:
            fh.write("completed")

        return directory

    def run(self, nprocs) -> None:
        """
        Main entry point for client code.
        Run relaxations using a multiprocessing Pool with nprocs processors.
        If nprocs is None or negative, the total number of CPUs in the system is used.
        """
        py_path = self.workdir / "analyze.py"
        print("Writing python script to analyze the results in:", py_path.name)

        with PythonScript(py_path) as script:
            script.add_text("""
def main():
    from abipy.tools.printing import print_dataframe
    from abipy.ml.relax_scanner import RelaxScannerAnalyzer
    rsa = RelaxScannerAnalyzer.from_topdir(".")
    #print_dataframe(rsa.df)
    #rsa.histplot()

    # Tolerances for pairs detection.
    ediff_tol, dist_tol = 1e-3, 3.5

    # Find pairs and save them to file
    rsa.pairs_enediff_dist(ediff_tol, dist_tol, neb_method=None)

    # Find pairs and compute static transition path.
    # NO NEB HERE, just linear interpolation between final and end points and energy calculations.
    #rsa.pairs_enediff_dist(ediff_tol, dist_tol, neb_method="static")

    # Find pairs and compute transition path with NEB.
    #rsa.pairs_enediff_dist(ediff_tol, dist_tol, neb_method="aseneb")
""").add_main()

        ntasks = self.nx * self.ny * self.nz
        nprocs = nprocs_for_ntasks(nprocs, ntasks, title="Begin relaxations")
        if nprocs == 1:
            # Sequential execution.
            self.run_start_count(0, -1)
        else:
            # Split ntasks across nprocs.
            args_list = []
            div, rest = divmod(ntasks, nprocs)
            for iproc in range(nprocs):
                start, stop = iproc * div, (iproc + 1) * div
                count = div
                if iproc == nprocs - 1:
                    stop += rest
                    count = div + rest
                args_list.append((self, start, count))

            with Pool(processes=nprocs) as pool:
                pool.map(_map_run_start_count, args_list)


def _map_run_start_count(args):
    """Function passed to pool.map."""
    scanner, start, count = args
    return scanner.run_start_count(start, count)


def _map_run_pair(kwargs):
    """Function passed to pool.map."""
    self = kwargs.pop("self")
    return self.run_pair(**kwargs)



class RelaxScannerAnalyzer:
    """
    Analyze the results produced by RelaxScanner.
    The object is usually constructed by calling `from_topdir`:

    Example:

        from abipy.ml.relax_scanner import RelaxScannerAnalyzer
        rsa = RelaxScannerAnalyzer.from_topdir(".")

        print_dataframe(rsa.df)
        rsa.histplot()
        rsa.pairs_enediff_dist(ediff_tol=1e-3, dist_tol=3.5, neb_method=None)
    """

    @classmethod
    def from_topdir(cls, topdir: Path) -> RelaxScannerAnalyzer:
        """
        Merge all entries starting from directory `topdir`.
        """
        topdir = Path(topdir)
        top = str(topdir)
        dirpaths = [name for name in os.listdir(top) if os.path.isdir(os.path.join(top, name))
                    and name.startswith("start_")]

        entries = []
        for dpath in dirpaths:
            pickle_path = topdir / Path(dpath) / "entries.pickle"
            completed_path = topdir / Path(dpath) / "COMPLETED"
            if not (pickle_path.exists() and completed_path.exists()): continue
            with open(pickle_path, "rb") as fh:
                for e in pickle.load(fh):
                    if e in entries: continue
                    entries.append(e)

        if not entries:
            raise ValueError(f"Entries is empty. Cannot find `entries.pickle` files in {topdir=}")

        # Reconstruct RelaxScanner instance from pickle file.
        scanner = RelaxScanner.pickle_load(topdir)
        return cls(entries, scanner)

    def __init__(self, entries: list[Entry], scanner: RelaxScanner, verbose: int = 0):
        self.entries = entries
        self.scanner = scanner
        self.verbose = verbose

    @property
    def workdir(self):
        return self.scanner.workdir

    #@property
    #def initial_structure(self):
    #    return self.scanner.initial_structure

    @lazy_property
    def df(self) -> pd.DataFrame:
        """
        Dataframe with the total energy in eV and the relaxed coords
        of the atomic site that has been relaxed.
        """
        dict_list = []
        entries = self.entries
        for entry in entries:
            site = entry.structure[entry.isite]
            x, y, z, fx, fy, fz = (*site.coords, *site.frac_coords)
            fmods = np.array([np.linalg.norm(force) for force in entry.forces])
            dict_list.append(dict(
                energy=entry.energy,
                xcart0=x, xcart1=y, xcart2=z,
                xred0=fx, xred1=fy, xred2=fz,
                fmin=fmods.min(), fmax=fmods.max(), fmean=fmods.mean(),
            ))

        df = pd.DataFrame(dict_list).sort_values(by="energy", ignore_index=True)

        # Add medata
        df.attrs['lattice_matrix'] = entries[0].structure.lattice.matrix
        df.attrs['structure'] = entries[0].structure
        return df

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        s = self.scanner.to_string(verbose=verbose)
        return s

    @lazy_property
    def lattice(self):
        return Lattice(self.df.attrs["lattice_matrix"])

    #def get_emin_emax(self, fact=0.001) -> tuple[float, float]:
    #    """
    #    Return min and max energy. The range is extended by .1% on each side.
    #    """
    #    emin, emax = self.df['energy'].min(), self.df['energy'].max()
    #    emin -= fact * abs(emin); emax += fact * abs(emax)
    #    return emin, emax

    def pairs_enediff_dist(self, ediff_tol=1e-3, dist_tol=3.5, neb_method=None, nprocs=-1) -> list[Pair]:
        """
        Find pairs (i.e. relaxed configurations) that differ in energy less than ediff_tol
        and with relaxed sites that are less than dist_tol Angstrom apart
        (minimum-image convention is applied).

        Args:
            ediff_tol: Energy difference in eV.
            dist_tol: Tolerance on site distance in Ang.

        Return: list of Pair objects.
        """
        def adiff_matrix(vec):
            """Return matrix A_ij with all the differences |vec_i - vec_j|."""
            x = np.reshape(vec, (len(vec), 1))
            return np.abs(x - x.transpose())

        energy = self.df["energy"].to_numpy(dtype=float)
        aediff_mat = adiff_matrix(energy)
        xreds = self.df[["xred0", "xred1", "xred2"]].to_numpy(dtype=float)

        # Find pairs.
        pairs = []
        inds = np.triu_indices_from(aediff_mat)
        for i, j in zip(inds[0], inds[1]):
            if aediff_mat[i,j] > ediff_tol or i == j: continue
            _, d2 = pbc_shortest_vectors(self.lattice, xreds[i], xreds[j], return_d2=True)
            dist = np.sqrt(float(d2))
            if dist > dist_tol: continue
            pair = Pair(index1=i, index2=j, ediff=energy[j] - energy[i], dist=dist,
                        frac_coords1=xreds[i], frac_coords2=xreds[j])
            pairs.append(pair)

        print(f"Found {len(pairs)} pair(s) with {ediff_tol=} eV and {dist_tol=} Ang.")

        if neb_method is None:
            # Build dataframe with pairs info.
            df = pd.DataFrame([pair.get_dict4pandas() for pair in pairs])
        else:
            # Compute the transition energy for each pair either by
            # performing NEB or single point calculations along a linear path connecting the two sites.
            nprocs = nprocs_for_ntasks(nprocs, len(pairs),
                                       title=f"Computing transition energies for each pair with {neb_method=}")

            if nprocs == 1:
                d_list = [self.run_pair(pair, neb_method=neb_method) for pair in pairs]
            else:
                kws_list = [dict(self=self, pair=pair, neb_method=neb_method) for pair in pairs]
                with Pool(processes=nprocs) as pool:
                    d_list = pool.map(_map_run_pair, kws_list)

            df = pd.DataFrame(d_list)
            print(df)

        df["ediff_tol"], df["dist_tol"] = ediff_tol, dist_tol
        df = df.sort_values(by=["ediff", "dist"], ignore_index=True)
        print_dataframe(df)
        path = self.workdir / f"{neb_method=}_data.csv".replace("'", "")
        print(f"Saving results to {path} in csv format.")
        df.to_csv(path)

        return pairs

    def run_pair(self, pair: Pair, neb_method="static", nimages=14, climb=False) -> dict:
        """
        Perform NEB calculation for the given pair. Return dictionary with results.
        NB: Contraints are enforces during the NEB

        Args:
            pair: Info on Pair.
            neb_method: One of the NEB methods implemented by ASE e.g. "aseneb"
                or "static" to compute total energies along a static path.
            nimages: Number of images
            climb: True to use climbing images.
        """
        # Get atoms with Fixed constraints.
        scanner = self.scanner
        delta_frac = pair.frac_coords2 - pair.frac_coords1
        initial_atoms = scanner.get_atoms_with_frac_coords(pair.frac_coords1)
        final_atoms = scanner.get_atoms_with_frac_coords(pair.frac_coords2)

        relax_mode = "no"
        pair_str = f"{pair.index1}-{pair.index2}"
        calc_builder = CalcBuilder(scanner.nn_name)

        # TODO: Check mic
        if neb_method == "static":
            # Just total energy calculations with a fixed path.
            calculators = [calc_builder.get_calculator() for i in range(nimages)]
            neb = make_ase_neb(initial_atoms, final_atoms, nimages,
                               calculators, "aseneb", climb,
                               method='linear', mic=False)

            #from abipy.ase.neb import interpolate
            #interpolate(images, mic=False, apply_constraint=False)

            # NB: It's not a NEB ML object but it provides a similar API.
            ml_neb = MlGsList(neb.images, calc_builder, self.verbose,
                              workdir=self.workdir / f"GSLIST/pair_{pair_str}")

        else:
            # Real NEB stuff
            ml_neb = MlNeb(initial_atoms, final_atoms,
                           nimages, neb_method, climb, scanner.optimizer_name,
                           relax_mode, scanner.fmax, scanner.pressure,
                           calc_builder, self.verbose,
                           workdir=self.workdir / f"NEB/pair_{pair_str}")

        if self.verbose: print(ml_neb.to_string(verbose=self.verbose))

        ml_neb.run()
        neb_data = ml_neb.read_neb_data()

        # Build out_data dict.
        out_data = pair.get_dict4pandas()

        # Add keys from neb_data.
        keys = ["barrier_with_fit", "energy_change_with_fit",
                "barrier_without_fit", "energy_change_without_fit"
               ]
        out_data.update({k: neb_data[k] for k in keys})
        out_data.update(dict(neb_method=neb_method, nimages=nimages, climb=climb))

        return out_data

    @add_fig_kwargs
    def histplot(self, ax=None, **kwargs):
        """
        Plot histogram to show energy distribution.
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        import seaborn as sns
        sns.histplot(self.df, x="energy", ax=ax)
        return fig


#def linear_path_with_extra_points(isite, initial, final, nimages, nextra, step) -> list[Atoms]:
#
#    # Get cart coords
#    p0 = initial[isite].position
#    p1 = final[isite].position
#
#    dvers_10 = (p1 - p0) / np.linalg.norm(p1 - p0)
#
#    from ase.neb import interpolate
#    first_segment = []
#    for i in range(nextra - 1):
#        new = initial.copy()
#        new[isite].position = p0 - i * step * dvers_10
#        first_segment.append(new)
#
#    images += [initial.copy() for i in range(nimages - 2)]
#
#    for i in range(nextra):
#        new = final.copy()
#        new[isite].position = p1 + i * step * dvers_10
#        images.append(new)
#
#    interpolate(images, mic=False, interpolate_cell=False, use_scaled_coord=False, apply_constraint=None)
#
#    return first_segment + images + last_segment



# coding: utf-8
"""
This module defines constants and conversion factors matching those present in abinit that can be used when
it is important to preserve consistency with the results produced by abinit.
"""
from __future__ import annotations

import numpy as np

# taken from abinit/10_defs/defs_basis.F90
# 1 Bohr, in Angstrom
Bohr_Ang = 0.52917720859
# 1 Angstrom in Bohr
Ang_Bohr = 1.0/Bohr_Ang
# 1 Hartree, in cm^-1
Ha_cmm1 = 219474.6313705
# 1 Hartree, in eV
Ha_eV = 27.21138386
Ha_to_eV = Ha_eV
Ha_to_meV = Ha_eV * 1000
# 1 eV in Hartree
eV_Ha = 1. / Ha_eV
# 1 eV in Rydberg
eV_Ry = 2 * eV_Ha
# 1 Hartree, in meV
Ha_meV = Ha_eV * 1000
# 1 Hartree, in Kelvin
Ha_K = 315774.65
# 1 eV, in Kelvin
eV_to_K = eV_Ha * Ha_K
# 1 Hartree, in THz
Ha_THz = 6579.683920722
# 1 Hartree, in s
Ha_s = Ha_THz * 1e12 * 2 * np.pi
# 1 eV, in THz
eV_to_THz = eV_Ha * Ha_THz
# 1 eV, in cm-1
eV_to_cm1 = 8065.5440044136285
# 1 Hartree, in J
Ha_J = 4.35974394e-18
# Planck constant and h/2pi in eV x s
h_eVs = 4.13566766225e-15
hbar_eVs = h_eVs / (2 * np.pi)
# minus the electron charge, in Coulomb
e_Cb = 1.602176487e-19
# Boltzmann constant in eV/K and Ha/K
kb_eVK = 8.617343e-5
kb_HaK = kb_eVK / Ha_eV
# 1 atomic mass unit, in electronic mass
amu_emass = 1.660538782e-27 / 9.10938215e-31
# 1 Ha/Bohr^3, in GPa
HaBohr3_GPa = Ha_eV / Bohr_Ang**3 * e_Cb * 1.0e+21
# 1 eV/A^3 to GPa
eVA3_GPa = 160.21766208
# 1 eV in seconds
eV_s = eV_to_THz*1e12 * 2*np.pi
# conversion factor for velocity between atomic units and SI
velocity_at_to_si = 2.1876912633e6

# per mole
Avogadro = 6.02214179e23
# 1 Ohm.cm in atomic units
Ohmcm = 2 * np.pi * Ha_THz * 10 / 9
eps0 = 1 / (4*np.pi*0.0000001*299792458.0**2)
AmuBohr2_Cm2 = e_Cb * 1.0e20 / (Bohr_Ang * Bohr_Ang)
# Inverse of fine structure constant
InvFineStruct = 137.035999679
# speed of light in SI
Sp_Lt_SI = 2.99792458e8
# speed of light in atomic units
Sp_Lt = Sp_Lt_SI / 2.1876912633e6
# Atomic unit of time, in seconds
Time_Sec = 2.418884326505e-17
# Tesla in a.u.
BField_Tesla = 4.254383e-6
# Debye unit in a.u.
dipole_moment_debye = 0.393430307


def phfactor_ev2units(units: str) -> float:
    """
    Return conversion factor eV --> units for phonons (case-insensitive)
    """
    d = {"ev": 1, "mev": 1000, "ha": eV_Ha,
         "cm-1": eV_to_cm1, 'cm^-1': eV_to_cm1,
         "thz": eV_to_THz,
         }
    try:
        return d[units.lower().strip()]
    except KeyError:
        raise KeyError('Value for units `{}` unknown\nPossible values are:\n {}'.format(units, list(d.keys())))


def phunit_tag(units: str, unicode=False) -> str:
    """
    Mainly used for phonons.
    Return latex string from ``units``.
    If unicode is True, replace Latex superscript with unitcode.
    """
    d = {"ev": "(eV)", "mev": "(meV)", "ha": '(Ha)',
         "cm-1": "(cm$^{-1}$)", 'cm^-1': "(cm$^{-1}$)", "thz": '(Thz)',
         }
    try:
        s = d[units.lower().strip()]
    except KeyError:
        raise KeyError('Value for units `{}` unknown\nPossible values are:\n {}'.format(units, list(d.keys())))

    if unicode:
        s = s.replace('$^{-1}$', '⁻¹')

    return s


def wlabel_from_units(units: str, unicode=False) -> str:
    """
    Return latex string for phonon frequencies in ``units``.
    """
    d = {'ev': 'Energy (eV)', 'mev': 'Energy (meV)', 'ha': 'Energy (Ha)',
         'cm-1': r'Frequency (cm$^{-1}$)',
         'cm^-1': r'Frequency (cm$^{-1}$)',
         'thz': r'Frequency (Thz)',
    }
    try:
        s = d[units.lower().strip()]
    except KeyError:
        raise KeyError('Value for units `{}` unknown\nPossible values are:\n {}'.format(units, list(d.keys())))

    if unicode:
        s = s.replace('$^{-1}$', '⁻¹')

    return s


def phdos_label_from_units(units: str, unicode=False) -> str:
    """
    Return latex string for phonon DOS values in ``units``.
    """
    d = {"ev": "(states/eV)", "mev": "(states/meV)", "ha": '(states/Ha)',
         "cm-1": "(states/cm$^{-1}$)", 'cm^-1': "(states/cm$^{-1}$)",
         "thz": '(states/Thz)',
        }
    try:
        s = d[units.lower().strip()]
    except KeyError:
        raise KeyError('Value for units `{}` unknown\nPossible values are:\n {}'.format(units, list(d.keys())))

    if unicode:
        s = s.replace('$^{-1}$', '⁻¹')

    return s


def s2itup(comp: str) -> tuple:
    """
    Convert string in the form ``xx``, ``xyz`` into tuple of two (three) indices
    that can be used to slice susceptibility tensors (numpy array).

    >>> assert s2itup("yy") == (1, 1)
    >>> assert s2itup("xyz") == (0, 1, 2)
    """
    d = {"x": 0, "y": 1, "z": 2}
    comp = str(comp).strip()
    if len(comp) == 2:
        return d[comp[0]], d[comp[1]]
    elif len(comp) == 3:
        return d[comp[0]], d[comp[1]], d[comp[2]]
    else:
        raise ValueError("Expecting component in the form `xy` or `xyz` but got `%s`" % comp)


def itup2s(t: tuple) -> str:
    """
    Convert tuple of 2 (3) integers into string in the form ``xx`` (``xyz``).
    Assume C-indexing e.g. 0 --> x

    >>> assert itup2s((0, 1)) == "xy"
    >>> assert itup2s((0, 1, 2)) == "xyz"
    """
    if not isinstance(t, tuple) and len(t) not in (2, 3):
        raise TypeError("Expecting tuple of len 2 or 3, got %s" % str(t))
    d = {0: "x", 1: "y", 2: "z"}
    return "".join(d[i] for i in t)


# Use same tolerances as Abinit Fortran version in m_occ.F90.
_maxFDarg = 500.0


@np.vectorize
def occ_fd(ee, kT, mu):
    r"""
    Fermi-Dirac statistic: 1 / [(exp((e - mu)/ KT) + 1]
    Note that occ_fs in [0, 1] so the spin factor is not included,
    unlike the occupations stored in ebands%occ.

    Args:
        ee: Single particle energy in eV
        kT: Value of K_Boltzmann x T in eV.
        mu: Chemical potential in eV.
    """
    ee_mu = ee - mu

    # 1 kelvin [K] = 3.16680853419133E-06 Hartree
    if kT > 1e-6 * Ha_eV:
        arg = ee_mu / kT
        if arg > _maxFDarg:
            occ_fd = 0.0
        elif arg < - _maxFDarg:
            occ_fd = 1.0
        else:
            occ_fd = 1.0 / (np.exp(arg) + 1.0)
    else:
        # Heaviside
        if ee_mu > 0.0:
            occ_fd = 0.0
        elif ee_mu < 0.0:
            occ_fd = 1.0
        else:
            occ_fd = 0.5

    return occ_fd


@np.vectorize
def occ_be(ee, kT, mu=0.0):
    r"""
    Bose-Einstein statistic: 1 / [(exp((e - mu)/ KT) - 1]

    Args:
      ee: Single particle energy in eV.
      kT: Value of K_Boltzmann x T in eV.
      mu: Chemical potential in eV (usually zero)
    """
    ee_mu = ee - mu

    # 1 kelvin [K] = 3.16680853419133E-06 Hartree
    if kT > 1e-12 * Ha_eV:
        arg = ee_mu / kT
        if arg > 1e-12 and arg < 600.0:
            occ_be = 1.0 / (np.exp(arg) - 1.0)
        else:
            occ_be = 0.0
    else:
        # No condensate for T --> 0
        occ_be = 0.0

    return occ_be



""""GUIs for structure."""
from __future__ import annotations 

import os
import io
import param
import panel as pn
import numpy as np
import panel.widgets as pnw
import bokeh.models.widgets as bkw

from abipy.core.structure import Structure
from abipy.panels.core import (AbipyParameterized, PanelWithStructure, dfc, mpl, ply,
    depends_on_btn_click, Loading, add_mp_rest_docstring)
from abipy.tools.decorators import Appender


add_inp_docstring = Appender("""
kprra (number of **k**-points per reciprocal atom) defines the **k**-mesh for electrons.
AbiPy automatically computes the variables [[ngkpt]], [[nshiftk]] and [[shiftk]] from kprra.

The pseudopotentials are taken from the [PseudoDojo project](http://www.pseudo-dojo.org/)
according to value of `XC type` and `Pseudos type` and recommended values for [[ecut]]
and [[pawecutdg]] and [[nband]] are automatically added to the input.

At the end of the page, there is a button to download a targz file with all the required input files.
""", indents=0)


def _make_targz_bytes(inp_or_multi, remove_dir=True):
    """
    Return bytesIO with the targz file containing input and pseudos.
    """
    targz_path = inp_or_multi.make_targz()
    output = io.BytesIO()
    with open(targz_path, "rb") as fh:
        output.write(fh.read())
    output.seek(0)

    return output


class StructurePanel(PanelWithStructure):
    """
    Panel with widgets to interact with an AbiPy Structure.
    """

    def __init__(self, structure: Structure, with_inputs: bool = True, **params):
        """
        Args:
            structure: |Structure| object.
            with_inputs: True if tabs for generating input files should be shown.
        """
        PanelWithStructure.__init__(self, structure=structure, **params)

        # Convert widgets.
        self.output_format = pnw.Select(name="format", value="abinit",
                                        options="abinit,cif,xsf,poscar,qe,siesta,wannier90,cssr,json".split(","))

        # Spglib widgets
        self.spglib_symprec = pnw.Spinner(name="symprec", value=0.01, start=0.0, end=None, step=0.01)
        self.spglib_angtol = pnw.Spinner(name="angtol", value=5, start=0.0, end=None, step=1)

        # Abisanitize widgets
        self.abisanitize_btn = pnw.Button(name="Run abisanitize", button_type='primary')
        self.select_primitive = pnw.Select(name='Select primitive',
                                           options=['primitive', 'primitive_standard', "no_primitive"])

        # K-path widgets
        self.kpath_format = pnw.Select(name="format", value="abinit", options=["abinit", "siesta", "wannier90"])
        self.line_density = pnw.Spinner(name="line density", value=10, step=5, start=0, end=None)
        self.plot_kpath = pnw.Checkbox(name='Plot k-path', value=False)

        # MP-match
        self.mp_match_btn = pnw.Button(name="Connect to Materials Project", button_type='primary')

        # MP-search
        #mp_search_btn = pnw.Button(name="Connect to Materials Project", button_type='primary')
        #mp_api_key

        # Widgets to control the generation of input files.
        self.kppra = pnw.Spinner(name="kppra", value=1000, step=500, start=0, end=None)
        self.smearing_type = pnw.Select(name="Smearing type", value=None,
                                        options=[None, "gaussian", "fermi_dirac"])
        self.tsmear = pnw.Spinner(name="tsmear (Ha)", value=0.01, step=0.002, start=0.0, end=None)

        # TODO: nspinor 2 should trigger a check on the table.
        from abipy.flowtk.psrepos import get_installed_repos_and_root
        installed_repos, repos_root = get_installed_repos_and_root()
        options = [repo.name for repo in installed_repos]
        default_repo = "ONCVPSP-PBE-SR-PDv0.4"
        default_repo = None if default_repo not in options else default_repo
        self.repos_name = pnw.Select(name="PP repos", value=default_repo, options=options)
        self.table_name = pnw.Select(name="Table name", value="standard", options=["standard", "stringent"])

        # widgets for GS input generator
        self.gs_type = pnw.Select(name="GS type", value="scf", options=["scf", "relax"])
        self.gs_input_btn = pnw.Button(name="Generate input", button_type='primary')

        # widgets for e-bands input generator.
        self.ebands_input_btn = pnw.Button(name="Generate input", button_type='primary')
        self.edos_kppra = pnw.Spinner(name="edos_kppra", value=0, step=500, start=0, end=None)

        # widgets for DFPT phonons input generator.
        self.ph_input_btn = pnw.Button(name="Generate input", button_type='primary')
        self.with_becs = pnw.Checkbox(name='BECS and epsilon_inf (semiconductors)', value=False)
        #self.dfpt_ngqpt = pnw.LiteralInput(name='ngqpt (python list)', value=[None, None, None], type=list)
        #self.dfpt_ngqpt = pnw.LiteralInput(name='ngqpt (python list)', value=[None, None, None], type=list)

        self.label2mode = {
            "unpolarized": 'unpolarized',
            "polarized": 'polarized',
            "non-collinear SOC with magnetism": "spinor",
            "non-collinear SOC, no magnetism": "spinor_nomag",
            "collinear anti-ferromagnetic": "afm",
        }

        self.spin_mode = pnw.Select(name="SpinMode", value="unpolarized", options=list(self.label2mode.keys()))

    @pn.depends("output_format.value")
    def convert(self):
        """Convert the input structure to one of the format selected by the user."""
        s = self.structure.convert(fmt=self.output_format.value)
        return self.html_with_clipboard_btn(f"<pre> {s} </pre>")

    @pn.depends("spglib_symprec.value", "spglib_angtol.value")
    def spglib_summary(self) -> pn.Row:
        """Call spglib to find space group symmetries and Wyckoff positions."""
        s = self.structure.spget_summary(symprec=self.spglib_symprec.value,
                                         angle_tolerance=self.spglib_angtol.value)
        return pn.Row(bkw.PreText(text=s, sizing_mode='stretch_width'))

    @depends_on_btn_click('abisanitize_btn')
    def on_abisanitize_btn(self) -> None:
        """
        Returns a new structure in which:

        - Structure is refined.
        - Reduced to primitive settings.
        - Lattice vectors are exchanged if the triple product is negative

        **symprec**:
            Symmetry precision used to refine the structure.

        **angle_tolerance**:
            Tolerance on angles.

        **primitive**:
            Returns most primitive structure found.

        **primitive_standard**:
            Whether to convert to a primitive cell using the standards defined in Setyawan, W., & Curtarolo, S. (2010).
            High-throughput electronic band structure calculations: Challenges and tools.
            Computational Materials Science, 49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010
        """
        primitive, primitive_standard = False, False
        if self.select_primitive.value in ('primitive', 'primitive_standard'):
            primitive = True
            primitive_standard = self.select_primitive.value == 'primitive_standard'
        else:
            assert self.select_primitive.value == "no_primitive"

        #print("primitive", primitive, "primitive_standard", primitive_standard)
        s = self.structure.abi_sanitize(symprec=self.spglib_symprec.value,
                                        angle_tolerance=self.spglib_angtol.value,
                                        primitive=primitive, primitive_standard=primitive_standard)

        return pn.Row(
                pn.Column("## Input Structure:", bkw.PreText(text=str(self.structure), sizing_mode='stretch_width')),
                pn.Column("## Sanitized:", bkw.PreText(text=str(s), sizing_mode='stretch_width')),
                sizing_mode='stretch_width')

    @pn.depends("kpath_format.value", "line_density.value")
    def get_kpath(self) -> pn.Column:
        """
        Generate high-symmetry k-path from input structure in the ABINIT format.
        """
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        s = self.structure.get_kpath_input_string(fmt=self.kpath_format.value,
                                                  line_density=self.line_density.value)
        ca(self.html_with_clipboard_btn(f"<pre> {s} </pre>"))

        if self.plot_kpath.value:
            # FIXME: I don't know why but these calls open in a new tab.
            ca("## Brillouin zone and **k**-path:")
            kpath_pane = ply(self.structure.plotly_bz(pmg_path=True, show=False), with_divider=False)
            df_kpts = dfc(self.structure.hsym_kpoints.get_highsym_datataframe(), with_divider=False)
            ca(pn.Row(kpath_pane, df_kpts))
            ca(pn.layout.Divider())

        return col

    def _get_pseudos_ecut_pawecutdg(self) -> tuple:
        #from abipy.data.hgh_pseudos import HGH_TABLE
        #return HGH_TABLE
        from abipy.flowtk.psrepos import get_repo_from_name
        repo = get_repo_from_name(self.repos_name.value)
        pseudos = repo.get_pseudos(self.table_name.value)
        ecut = 0
        pawecutdg = 0

        for p in pseudos:
            hint = p.hint_for_accuracy(accuracy="normal")
            ecut = max(ecut, hint.ecut)
            pawecutdg = max(pawecutdg, hint.pawecutdg)

        return pseudos, ecut, pawecutdg

    def _get_smearing(self):
        smearing = None
        if self.smearing_type.value is not None:
            smearing = "%s:%s Ha" % (self.smearing_type.value, self.tsmear.value)
        return smearing

    def _finalize(self, inp_or_multi, header=None, validate=False):

        # TODO chksymbreak should be set to 0 else Abinit may stop.
        if validate:
            if inp_or_multi.is_input:
                this = inp_or_multi.deepcopy()
                this.pop_vars("pseudos")
                v = this.abivalidate()
                v_list = [v]
            else:
                this = inp_or_multi.deepcopy()
                this.pop_vars("pseudos")
                v_list = this.abivalidate()

            count = 0
            for v in v_list:
                if v.retcode != 0:
                    count += 1
                    print(v.stderr_file.read())
                    print(v.log_file.read())
            if count != 0: raise RuntimeError("")

        # Here we replace the absolute paths with their basenames
        # so that we can create a targz file with the pseudos
        # We do it at this level because absolute paths are needed for executing abinit.
        inp_or_multi.set_vars(pseudos='"%s"' % ", ".join(p.basename for p in inp_or_multi.pseudos))

        html = self.html_with_clipboard_btn(inp_or_multi._repr_html_())

        pseudos = inp_or_multi.pseudos
        nspinor = inp_or_multi.get("nspinor", 1)
        if nspinor == 2 and any(not p.supports_soc for p in pseudos):
            raise RuntimeError("The pseudopotential table selected does not support calculations with SOC!")

        def download_input():
            return _make_targz_bytes(inp_or_multi)

        file_download = pnw.FileDownload(filename="input.tar.gz", callback=download_input)

        alert = pn.pane.Alert( """
This input file has been automatically generated by AbiPy.
Several input parameters have **default values** that might not be suitable for your particular calculation.
Please check the input variables and modify them according to your needs.

Also note that running big calculations with lot of datasets is not the most efficient approach.
Examples of AbiPy scripts to automate calculations without datasets are available
[at this page](https://abinit.github.io/abipy/flow_gallery/index.html).
""", alert_type="danger")

        items = [html, file_download, pn.layout.Divider(), alert]
        if header: items.insert(0, header)

        return pn.Column(*items, sizing_mode="stretch_width")

    def get_gs_input(self) -> AbinitInput:
        """
        Return an AbinitInput for GS calculation from the parameters selected via the widgets.
        """
        from abipy.abio.factories import gs_input
        pseudos, ecut, pawecutdg = self._get_pseudos_ecut_pawecutdg()

        gs_inp = gs_input(structure=self.structure,
                          pseudos=pseudos,
                          kppa=self.kppra.value,
                          ecut=ecut,
                          pawecutdg=pawecutdg,
                          spin_mode=self.label2mode[self.spin_mode.value],
                          smearing=self._get_smearing(),
                          charge=0.0,
                         )

        gs_inp.set_mnemonics(False)
        gs_inp.pop_vars(("charge", "chksymbreak"))

        # TODO: ecut, ....
        #gs_inp.set_vars(#ecut="??  # depends on pseudos",
        #                #nband="?? # depends on pseudos",
        #                pseudos='"%s"' % ", ".join(p.basename for p in gs_inp.pseudos),
        #                )

        return gs_inp

    @depends_on_btn_click('gs_input_btn')
    @add_inp_docstring
    def on_gs_input_btn(self):
        """
        This Tab provides widgets to generate a minimalistic ABINIT input file
        for ground-state calculations or structural relaxations.
        In the case of relaxation runs, the following variables are automatically added:

        [[optcell]] = 2, [[ionmov]] = 2, [[ecutsm]] = 0.5 and [[dilatmx]] = 1.05.
        """
        gs_inp = self.get_gs_input()

        if self.gs_type.value == "relax":
            gs_inp.set_vars(optcell=2, ionmov=2, ecutsm=0.5, dilatmx=1.05)

        return self._finalize(gs_inp)

    @depends_on_btn_click('ebands_input_btn')
    @add_inp_docstring
    def on_ebands_input_btn(self):
        """
        This Tab provides widgets to generate a minimalistic ABINIT input file
        for band structure calculations.
        The first dataset performs a GS run to obtain the density file.
        The second dataset reads the DEN file and performs a NSCF calculation on
        a high-symmetry **k**-path generated automatically according to the input structure.
        [[ndivsm]] and [[kptbounds]]

        Optionally, one can add a third dataset to compute the e-DOS via a NSCF run on a
        much denser **k**-mesh.
        """
        from abipy.abio.factories import ebands_input

        dos_kppa = self.edos_kppra.value
        if dos_kppa == 0.0: dos_kppa = None
        pseudos, ecut, pawecutdg = self._get_pseudos_ecut_pawecutdg()

        multi = ebands_input(structure=self.structure,
                             pseudos=pseudos,
                             kppa=self.kppra.value,
                             nscf_nband=None,
                             ndivsm=10,
                             ecut=ecut,
                             pawecutdg=pawecutdg,
                             scf_nband=None,
                             spin_mode=self.label2mode[self.spin_mode.value],
                             smearing=self._get_smearing(),
                             charge=0.0,
                             dos_kppa=dos_kppa,
        )

        # Add getwfk variables.
        for inp in multi[1:]:
            inp["getwfk"] = 1

        multi.pop_vars(("charge", "chksymbreak"))
        #multi.set_vars(#ecut="??  # depends on pseudos",
        #               #nband="?? # depends on pseudos",
        #               )

        return self._finalize(multi)

    @depends_on_btn_click('ph_input_btn')
    @add_inp_docstring
    def on_ph_input_btn(self):
        """
        This Tab provides widgets to generate a minimalistic ABINIT input file
        for the DFPT computation of phonons, Born effective charges (BECS)
        and macroscopic dieletric tensors.
        """
        from abipy.abio.factories import phonons_from_gsinput

        gs_inp = self.get_gs_input()
        ngkpt = np.array(gs_inp["ngkpt"])

        # TODO Find an easy way to specify the q-mesh.
        # Use ngqpt a la Samuel?
        ph_ngqpt = [2, 2, 2]
        #ph_nqpt = qscale * ngkpt
        #any(ngkp % ph_ngqpt != 0):

        qpoints = gs_inp.abiget_ibz(ngkpt=ph_ngqpt, shiftk=[0, 0, 0], kptopt=1).points

        ndtset = 1 + len(qpoints)
        qstart = 1
        with_becs = self.with_becs.value
        if with_becs:
            # Account for the DDK (2nd dataset)
            ndtset += 1
            qstart = 2

        multi = gs_inp.replicate(ndtset)

        if with_becs:
            ddk_input = multi[qstart - 1]
            ddk_input.set_vars(
                    rfelfd=2,          # only the derivative of ground-state wavefunctions with respect to k
                    rfdir=[1, 1, 1],
                    nqpt=1,
                    qpt=(0, 0, 0),
                    kptopt=2,         # 2 to take into account time-reversal symmetry.
                    iscf=-3,          # The d/dk perturbation must be treated in a non-self-consistent way
                    getwfk=1,
                    comment="Response Function calculation: d/dk",
            )
            ddk_input.pop_tolerances()
            ddk_input.set_vars(tolwfr=1.0e-22)

        for qpt, ph_input in zip(qpoints, multi[qstart:]):
            is_gamma = np.sum(qpt ** 2) < 1e-12
            #if with_becs and is_gamma: continue
            ph_input.set_vars(
                    kptopt=2 if is_gamma else 3,  # use time-reversal if Gamma
                    rfphon=1,                     # Will consider phonon-type perturbation
                    nqpt=1,                       # One wavevector is to be considered
                    qpt=qpt,                      # q-wavevector.
                    rfatpol=[1, len(gs_inp.structure)],
                    rfdir=[1, 1, 1],
                    getwfk=1,
                    rfelfd=3 if (with_becs and is_gamma) else None,
                    getddk=2 if (with_becs and is_gamma) else None,
                    prtwf=-1,
                    comment="Input file for PH calculation with DFPT.",
            )
            ph_input.pop_tolerances()
            ph_input.set_vars(tolvrs=1.0e-10)

        header = f"""
## BZ sampling

- k-mesh for electrons: {ngkpt}
- qmesh for phonons: {ph_ngqpt}
"""

        #multi = phonons_from_gsinput(gs_inp, ph_ngqpt=None, qpoints=None, with_ddk=True,
        #                             with_dde=True, with_bec=False,
        #                             ph_tol=None, ddk_tol=None, dde_tol=None, wfq_tol=None,
        #                             qpoints_to_skip=None, manager=None)

        # Add getwfk variables.
        #for inp in multi[1:]:
        #    inp["getwfk"] = 1

        #multi.pop_vars(("charge", "chksymbreak"))
        #multi.set_vars(#ecut="??  # depends on pseudos",
        #               #nband="?? # depends on pseudos",
        #               pseudos='"%s"' % ", ".join(p.basename for p in multi.pseudos),
        #               )

        return self._finalize(multi, header=header)

    @depends_on_btn_click('mp_match_btn', show_shared_wdg_warning=False)
    @add_mp_rest_docstring
    def on_mp_match_btn(self):
        """
        Match your structure with the ones available in the MP database.
        Produce table with lattice parameters and space group info.
        """
        mp = self.structure.mp_match()
        if not mp.structures:
            return pn.Column("## No structure found in the MP database")

        return pn.Column(dfc(mp.lattice_dataframe, transpose=True),
                         sizing_mode='stretch_width')

    #@depends_on_btn_click('mp_search_btn')
    #def on_mp_search_btn(self):
    #    from abipy.core.structure import mp_search
    #    chemsys_formula_id = self.stucture.formula
    #    mp = mp_search(chemsys_formula_id, api_key=None, endpoint=None, final=True)
    #    if not mp.structures:
    #        raise RuntimeError("No structure found in MP database")

    #    return pn.Column(dfc(mp.lattice_dataframe), sizing_mode='stretch_width')

    def get_panel(self, with_inputs=True, as_dict=False, **kwargs):
        """
        Build panel with widgets to interact with the structure either in a notebook or in a bokeh app.

        Args:
            with_inputs: True if tabs for generating input files should be shown.
        """
        d = dict()

        d["Summary"] = self.get_summary_view_for_abiobj(self.structure)
        d["Viewer"] = self.get_structure_view()
        d["Spglib"] = pn.Row(
            self.pws_col(['## Spglib options',
                          "spglib_symprec", "spglib_angtol",
                        ]),
            self.spglib_summary
        )
        d["AbiSanitize"] = pn.Row(
            self.pws_col(['## Spglib options',
                          "spglib_symprec", "spglib_angtol", "select_primitive", "abisanitize_btn",
                          pn.layout.Divider(),
                        ]),
            self.on_abisanitize_btn
        )
        d["Kpath"] = pn.Row(
            self.pws_col(['## K-path options',
                          "kpath_format", "line_density", "plot_kpath",
                         ]),
            self.get_kpath
        )
        d["Convert"] = pn.Row(
            self.pws_col(["## Convert structure", "output_format",
                         ]),
            self.convert
        )


        if with_inputs:
            # Add tabs to generate inputs from structure.
            d["GS-input"] = pn.Row(
                self.pws_col(['## Generate GS input',
                              "gs_type", "spin_mode", "kppra", "smearing_type", "tsmear",
                              "repos_name", "table_name",
                              "gs_input_btn",
                             ]),
                self.on_gs_input_btn
            )

            d["Ebands-input"] = pn.Row(
                self.pws_col(['## Generate Ebands input',
                              "spin_mode", "kppra", "edos_kppra", "smearing_type", "tsmear",
                              "repos_name", "table_name",
                              "ebands_input_btn",
                            ]),
                self.on_ebands_input_btn
            )
            d["PH-input"] = pn.Row(
                self.pws_col(['## Generate phonon input',
                              "spin_mode", "kppra", "smearing_type", "tsmear",
                              "with_becs",
                              "repos_name", "table_name",
                              "ph_input_btn",
                            ]),
                self.on_ph_input_btn
            )

        d["MP-match"] = pn.Column(self.on_mp_match_btn,
                                  pn.layout.Divider(),
                                  pn.Row(self.mp_match_btn, align="center"),
                                  sizing_mode="stretch_width")

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


class InputFileGenerator(AbipyParameterized):

    abi_sanitize = param.Boolean(True, doc="Sanitize structure")

    info_str = """
Generate ABINIT input files for performing basic:

   - GS calculations
   - NSCF band structure calculations
   - DFPT calculations for phonons including Born effective charges and dielectric constant

starting from a crystalline structure provided by the user either via an external file
or through the Materials Project identifier (*mp-id*).
"""

    def __init__(self, **params):

        super().__init__(**params)

        # Spglib widgets
        #self.spglib_symprec = pnw.Spinner(name="symprec", value=0.01, start=0.0, end=None, step=0.01)
        #self.spglib_angtol = pnw.Spinner(name="angtol", value=5, start=0.0, end=None, step=1)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md,
                                   self.get_alert_data_transfer(),
                                   sizing_mode="stretch_width")

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")

        self.mpid_input = pnw.TextInput(name='mp-id', placeholder='Enter e.g. mp-149 for Silicon and press ⏎')
        self.mpid_input.param.watch(self.on_mpid_input, "value")
        self.mpid_err_wdg = pn.pane.Markdown("")

    def _set_structure(self, structure):
        if self.abi_sanitize:
            structure = structure.abi_sanitize(symprec=1e-3, angle_tolerance=5,
                                               primitive=True, primitive_standard=False)
        self.input_structure = structure

    def on_file_input(self, event):
        self.mpid_err_wdg.object = ""
        #print("filename", self.file_input.filename)
        if self.file_input.value is None: return None
        #print("value", self.file_input.value)
        import tempfile
        workdir = tempfile.mkdtemp()

        fd, tmp_path = tempfile.mkstemp(suffix=self.file_input.filename)
        with open(tmp_path, "wb") as fh:
            fh.write(self.file_input.value)

        self._set_structure(Structure.from_file(tmp_path))

        os.remove(tmp_path)
        self.update_main_area()

    def on_mpid_input(self, event):
        with Loading(self.mpid_input, err_wdg=self.mpid_err_wdg):
            self._set_structure(Structure.from_mpid(self.mpid_input.value))

        self.update_main_area()

    def update_main_area(self):
        with Loading(self.main_area):
            d = self.input_structure.get_panel(as_dict=True)
            d = {k: d[k] for k in ("GS-input", "Ebands-input", "PH-input", "Summary")}
            tabs = pn.Tabs(*d.items())
            self.main_area.objects = [tabs]

    def get_panel(self):
        col = pn.Column(
            "## Upload (or drag & drop) **any file** with a structure (*.nc*, *.abi*, *.cif*, *.xsf*, *POSCAR*):",
            self.get_fileinput_section(self.file_input),
            "## or get the structure from the [Materials Project](https://materialsproject.org/) database:",
            pn.Row(self.mpid_input, pn.Column(self.mpid_err_wdg)),
            "## Invoke abi_sanitize to refine structure",
            self.param.abi_sanitize,
        sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")
        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="Input File Generator", **kwds)



#!/usr/bin/env python
r"""
Unit cell
=========

This example shows how to display the unit cell with matplotlib.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

#%%
# Extract structure from the netcdf file:

with abiopen(abidata.ref_file("sio2_kpath_GSR.nc")) as gsr:
    structure = gsr.structure

#%%
# To visualize the structure with matplotlib, use:
structure.plot(color_scheme="Jmol")

#%%
# To wrap sites into first unit cell, use:

# sphinx_gallery_thumbnail_number = 2
structure.plot(to_unit_cell=True)



"""
"""
from __future__ import annotations

import abc
import os
import json
import numpy as np

#from typing import
#from monty.io import FileLock
#from pymatgen.core.xcfunc import XcFunc
from abipy.core.structure import Structure
#from abipy.abio.factories import ion_ioncell_relax_input
#from abipy.flowtk.abiobjects import SpinMode, Smearing, KSampling, RelaxationMethod
#from abipy.flowtk.tasks import RelaxTask
#from abipy.flowtk.works import Work, RelaxWork, PhononWork
#from pseudo_dojo.core.dojoreport import DojoReport, dojo_dfact_esults, dojo_gbrv_results
#from pseudo_dojo.refdata.gbrv import gbrv_database
#from pseudo_dojo.refdata.deltafactor import df_database
#from pseudo_dojo.refdata.lantanides.database import raren_database

from abipy.electrons.gsr import GsrRobot
from abipy.abio.inputs import AbinitInput
from abipy.abio.factories import gs_input
#from abipy.flowtk.pseudos import Pseudo
from abipy.flowtk.works import Work
from abipy.flowtk.flows import Flow


#class PseudoTestsFlow(Flow):

    #@classmethod
    #def from_pseudo_filepath(cls, pseudo_filepath, ecut_list, workdir):
    #    flow = cls(workdir=workdir)
    #    flow.pseudo_filepath = os.path.abspath(pseudo_filepath)
    #    scf_input = _build_scf_input_from_pseudo(pseudo)
    #    work = GsEcutConvWork.from_scf_input(cls, scf_input, pseudo, ecut_list)
    #    flow.register_work(work)

    #    return flow

    #def on_all_ok(self):
    #    """
    #    This method is called when all tasks have reached S_OK.
    #    It reads the energies and the volumes from the GSR file
    #    """
    #    for work in self:
    #        with open(work.outdir.path_in("ecut_conv.json"), "wt") as fh:
    #            data[work.KEY] = json.read(fh)

    #    with open(self.outdir.path_in("ecut_conv.json"), "wt") as fh:
    #        json.dump(data, fh, indent=4, sort_keys=True)

#class _PseudoFlow(Flow):
#    """Base class for Flow."""


class _PseudoWork(Work):
    """Base class for Works."""
    #__metaclass__ = abc.ABCMeta

    #@abc.abstractproperty
    #def dojo_pseudo(self):
    #    """:class:`Pseudo` object"""

    #@abc.abstractproperty
    #def dojo_trial(self):
    #    """String identifying the DOJO trial. Used to write results in the DOJO_REPORT."""

    #@property
    #def djrepo_path(self):
    #    """Path to the djrepo file."""
    #    root, ext = os.path.splitext(self.dojo_pseudo.filepath)
    #    return root + ".djrepo"

    #def add_entry_to_dojoreport(self, entry, overwrite_data=False, pop_trial=False):
    #    """
    #    Write/update the DOJO_REPORT section of the pseudopotential.
    #    Important parameters such as the name of the dojo_trial and the energy cutoff
    #    are provided by the sub-class.
    #    Client code is responsible for preparing the dictionary with the data.

    #    Args:
    #        entry: Dictionary with results.
    #        overwrite_data: If False, the routine raises an exception if this entry is
    #            already filled.
    #        pop_trial: True if the trial should be removed before adding the new entry.
    #    """
    #    djrepo = self.djrepo_path
    #    self.history.info("Writing dojreport data to %s" % djrepo)

    #    # Update file content with Filelock.
    #    with FileLock(djrepo):
    #        # Read report from file.
    #        file_report = DojoReport.from_file(djrepo)

    #        # Create new entry if not already there
    #        dojo_trial = self.dojo_trial

    #        if pop_trial:
    #            file_report.pop(dojo_trial, None)

    #        if dojo_trial not in file_report:
    #            file_report[dojo_trial] = {}

    #        # Convert float to string with 1 decimal digit.
    #        dojo_ecut = "%.1f" % self.ecut

    #        # Check that we are not going to overwrite data.
    #        if dojo_ecut in file_report[dojo_trial]:
    #            if not overwrite_data:
    #                raise RuntimeError("dojo_ecut %s already exists in %s. Cannot overwrite data" %
    #                        (dojo_ecut, dojo_trial))
    #            else:
    #                file_report[dojo_trial].pop(dojo_ecut)

    #        # Update file_report by adding the new entry and write new file
    #        file_report[dojo_trial][dojo_ecut] = entry

    #        # Write new dojo report and update the pseudo attribute
    #        file_report.json_write()
    #        self._pseudo.dojo_report = file_report


class GsEcutConvWork(_PseudoWork):
    """
    This work computes GS properties for different value of ecut
    and save the results in the outdata directory of the Work.
    """

    KEY = "gs_vs_ecut"

    @classmethod
    def from_pseudo(cls, pseudo, ecut_list: list, a=4) -> GsEcutConvWork:
        """
        High-level constructor to build the work from a pseudo and a list of ecut energies in Ha.
        """
        species = [pseudo.symbol]
        structure = Structure.fcc(a, species, primitive=True, units="ang")

        #scf_input = AbinitInput(structure, pseudo)
        scf_input = gs_input(structure, pseudo,
                 kppa=None, ecut=None, pawecutdg=None, scf_nband=None, accuracy="normal", spin_mode="unpolarized",
                 smearing="fermi_dirac:0.1 eV", charge=0.0, scf_algorithm=None)

        # Some optimization.
        scf_input.set_vars(nstep=50, prtwf=-1)

        return cls.from_scf_input(scf_input, ecut_list)

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput, ecut_list: list) -> GsEcutConvWork:
        """
        Low-level constructor to build the work from a template and a list of ecut energies.
        """
        work = cls()
        for ecut in ecut_list:
            work.register_scf_task(scf_input.new_with_vars(ecut=ecut))

        return work

    def on_all_ok(self):
        """
        This method is called when all tasks in GsEcutWork have reached S_OK.
        It reads the energies and the volumes from the GSR file
        and produces the `gs_vs_ecut.json` file in the outdir of the work.
        """
        with GsrRobot.from_work(self) as gsr_robot:
            df = gsr_robot.get_dataframe(with_geo=False)
            df.to_excel(self.outdir.path_in("gs_vs_ecut.xlsx"))

            with gsr_robot.get_pyscript(self.outdir.path_in("gsr_robot.py")) as script:
                script.add_text("""
# Set convergence criteria for absolute convergence in ytols_dict.
# NB: Energies are in eV, pressure in GPa.

ytols_dict = dict(energy_per_atom=1e-3, pressure=1e-2)
ca = robot.get_convergence_analyzer("ecut", ytols_dict)
print(ca)
ca.plot()
""")

        return super().on_all_ok()


#class GsEcutConvFlowPseudos(Flow):
#    """
#    """
#
#    @classmethod
#    def from_scf_input(cls, scf_input, pseudos_to_test, ecut_list, workdir=None):
#        flow = cls(workdir=workdir)
#        for pseudo in pseudos_to_test:
#            flow.append_work(GsEcutConvWork.from_scf_input(scf_input, pseudo, ecut_list))
#
#        return flow
#
#    def on_all_ok(self):
#        """
#        This method is called when all works in the Flow have reached S_OK.
#        """
#        for work in self:
#            with open(work.outdir.path_in("gs_vs_ecut.json"), "wt") as fh:
#                data = json.load(fh)



class PhEcutConvWork(_PseudoWork):
    """
    This work computes DFPT phonons for different ecut values
    and save the results in the outdata directory of the Work.
    """

    KEY = "phonons_vs_ecut"

    def from_scf_input(cls, scf_input, pseudo, ecut_list, workdir=None):
        work = cls(workdir=workdir)
        work.natom = len(scf_input.structure)
        work.ecut_list = list(ecut_list)
        #for ecut in self.ecut_list:
        #    work.register_scf_task(scf_input.new_with_vars(ecut=ecut))

        return work

    def on_all_ok(self):
        """
        This method is called when all tasks have reached S_OK.
        It reads the energies and the volumes from the GSR file
        """
        energy_per_atom_ev = []
        #pressure_gpa = []
        for task in self:
            with task.open_ddb() as ddb:
                print(ddb)

        data = dict(
            natom=self.natom,
            #ecut_list=self.ecut_list,
            #energy_per_atom_ev=energy_per_atom_ev,
            #pressure_gpa=pressure_gpa,
        )

        with open(self.outdir.path_in("ecut_conv.json"), "wt") as fh:
            json.dump(data, fh, indent=4, sort_keys=True)



#!/usr/bin/env python
from __future__ import print_function, unicode_literals, division, absolute_import
import sys

from abipy.data import AbinitFilesGenerator


class MyGenerator(AbinitFilesGenerator):
    """This class generates the output files used in the unit tests and in the examples."""
    # Subclasses must define the following class attributes:
    # List of pseudos (basenames) in abipy/data/pseudos
    pseudos = ["12mg.pspnc", "B.psp8"]
    #pseudos = ["B.psp8", "12mg.pspnc"]


    # Mapping old_name --> new_name for the output files that must be preserved.
    files_to_save = {
        "out_DS1_FATBANDS.nc": "mgb2_kmesh181818_FATBANDS.nc",
        "out_DS2_FATBANDS.nc": "mgb2_kpath_FATBANDS.nc",
    }


if __name__ == "__main__":
    sys.exit(MyGenerator().run())



"""
pydantic models storing results of ground-state calculations.
"""
from __future__ import annotations

from pydantic import Field
from typing import List
from abipy.electrons.gsr import GsrFile
from abipy.electrons.ebands import ElectronBands
from abipy.dynamics.hist import HistFile
from .base_models import AbipyModel, GfsFileDesc, GfsDesc, MongoConnector
#from .structure_models import StructureData


class _ModelWithGsr(AbipyModel):
    """
    This is the base class that defines the dimensions and the results of an "abstract"
    GS calculations i.e. an Abinit computation that produces a GSR file.

    This class is not supposed to be instanciated directly.
    It just contains Fields that are common to specialized subclasses corresponding to Scf/NSCF/Relax runs.
    """

    ebands: ElectronBands = Field(..., description="Electronic bands.")

    nkpt: int = Field(..., description="Number of k-points in ebands")

    nsppol: int = Field(..., description="Number of independent spins in WFs")

    nspinor: int = Field(..., description="Number of spinor components in WFs")

    nspden: int = Field(..., description="Number of independent spin components in density")

    fermie_ev: float = Field(..., description="Fermi energy in eV.")

    is_metal: bool = Field(..., description="True if system is metallic")

    fundamental_gap_ev: float = Field(None, description="Fundamental band gap in eV (min over spins, if any).")

    direct_gap_ev: float = Field(None, description="Direct band gap in eV (min over spins, if any).")

    ebands_gfsd: GfsDesc = Field(None, description="Pointer to the ElectronBands object in GridFS")

    gsr_gfsd: GfsFileDesc = Field(None, description="Pointer to the GSR.nc file stored in GridFS. "
                                                    "None if the GSR.nc is not stored")

    @classmethod
    def from_gsr_filepath(cls, gsr_filepath: str, mng_connector: MongoConnector, with_gsr: bool):
        """
        Fill the model from a GSR filepath.
        """
        with GsrFile(gsr_filepath) as gsr:
            return cls.from_gsr(gsr, mng_connector, with_gsr)

    @classmethod
    def get_base_data(cls, gsr: GsrFile, mng_connector: MongoConnector, with_gsr: bool) -> dict:

        #attrs = [
        #    "ecut", "pawecutdg",
        #    "tsmear", "nkpt",
        #    "nsppol", "nspinor", "nspden",
        #]

        fundamental_gap_ev, direct_gap_ev = 0.0, 0.0
        if not gsr.ebands.is_metal:
            fundamental_gap_ev = min(gap.energy for gap in gsr.ebands.fundamental_gaps)
            direct_gap_ev = min(gap.energy for gap in gsr.ebands.direct_gaps)

        data = dict(
            nkpt=len(gsr.ebands.kpoints),
            nsppol=gsr.nsppol,
            nspinor=gsr.nspinor,
            nspden=gsr.nspden,
            ebands=gsr.ebands,
            ebands_gfsd=mng_connector.gfs_put_mson_obj(gsr.ebands),
            is_metal=gsr.ebands.is_metal,
            fermie_ev=gsr.ebands.fermie,
            fundamental_gap_ev=fundamental_gap_ev,
            direct_gap_ev=direct_gap_ev,
        )

        if with_gsr:
            # Add the GSR file to GridFS.
            data["gsr_gfsd"] = mng_connector.gfs_put_filepath(gsr.filepath)

        return data


class NscfData(_ModelWithGsr):
    """
    Model for NSCF calculations.
    """

    #dos_ef
    #dos_values

    @classmethod
    def from_gsr(cls, gsr: GsrFile, mng_connector: MongoConnector, with_gsr: bool) -> NscfData:
        """
        Fill the model from a |GsrFile|.
        """
        if gsr.is_scf_run:
            raise RuntimeError("Expecting a GSR file produced by a NSCF calculation.")

        data = cls.get_base_data(gsr, mng_connector, with_gsr)

        return cls(**data)


class ScfData(_ModelWithGsr):
    """
    Model for SCF calculations with k-points in the IBZ. This kind of sampling allows one
    to compute energies, forces, stress tensor and collinear magnetization.
    """

    max_force_ev_over_ang: float = Field(...,
        description="Max absolute Cartesian force in eV/Ang. None if forces are not available")

    abs_pressure_gpa: float = Field(..., description="Pressure in GPa. Absolute value")

    pressure_gpa: float = Field(..., description="Pressure in GPa. NB: Value with sign!")

    energy: float = Field(..., description="The total DFT energy in eV")

    energy_per_atom: float = Field(..., description="The DFT energy per atom in eV")

    #forces: List[Vector3D] = Field(
    #    [], description="Forces on atoms from the last calculation"
    #)

    #cart_stress_tensor_gpa: MatrixLike = Field(None, description="Cartesian stress Tensor in GPA")

    #cart_forces_ev_over_ang: MatrixLike = Field(None, description="Cartesian forces in eV ang^-1")

    #cart_stress: Matrix3D = Field(...
    #    [], description="Stress on the unitcell from the last calculation"
    #)

    collinear_mag: float = Field(None,
        description="Total collinear magnetization in Bohr magneton as the difference "
                    "between the spin up and spin down densities. None if undefined e.g. nspinor 2")

    @classmethod
    def from_gsr(cls, gsr: GsrFile, mng_connector: MongoConnector, with_gsr: bool) -> ScfData:
        """
        Fill the model from a |GsrFile|
        """
        if not gsr.is_scf_run:
            raise RuntimeError("Expecting a GSR file produced by a SCF calculation")

        data = cls.get_base_data(gsr, mng_connector, with_gsr)

        # TODO: Use "cartesian_forces_eV/Ang" and get rid of ArrayWithUnits
        #gsr.ebands.structure.remove_site_property("cartesian_forces")
        #data = dict(
        #    ebands=gsr.ebands,
        #    ebands_gfsd=mng_connector.gfs_put_mson_obj(gsr.ebands),
        #)

        # Add entries that are available only if SCF run.
        try:
            collinear_mag = gsr.ebands.get_collinear_mag()
        except ValueError:
            collinear_mag = None

        data.update(dict(
            pressure_gpa=gsr.pressure,
            abs_pressure_gpa=abs(gsr.pressure),
            max_force_ev_over_ang=gsr.max_force,
            energy=float(gsr.energy),
            energy_per_atom=float(gsr.energy_per_atom),
            #cart_stress_tensor_gpa=??,
            #cart_forces_ev_over_ang=??.
            collinear_mag=collinear_mag,
        ))

        return cls(**data)


class RelaxData(ScfData):
    """
    Model for structural relaxations.
    Store energy, forces, stress tensor, Fermi level, band gaps for the final configuration
    inherited from ScfData as well as additional lists with the evolution of the most important
    physical parameters at each relaxation step for plotting purposes.
    """

    #final_structure_data = StructureData = Field(..., description="Data computed from the final structure")

    num_steps: int = Field(..., description="Number of relaxation steps.")

    #etotal_ev_hist: List[float] = Field(..., description="Energies in eV at the different relaxation steps")

    #pressure_gpa_hist: List[float] = Field(..., description="Pressure in GPa at the different relaxation steps")

    #angles_hist: List[List[float]] = Field(...,
    #                                       description="Lattice angles in degrees at the different relaxation steps")

    #lenghts_ang_hist: List[List[float]] = Field(...,
    #                                            description="Lattice lengths in Ang at the different relaxation steps")

    #volume_ang3_hist: List[float] = Field(...,
    #                                      description="Unit cell volume in Ang**3 at the different relaxation steps")


    hist_gfsd: GfsFileDesc = Field(None, description="Pointer to the HIST.nc stored in GridFS. None if not stored")

    @classmethod
    def from_gsr(cls, gsr: GsrFile, mng_connector: MongoConnector, with_gsr: bool) -> NscfData:
        raise RuntimeError("For {cls.__name__}, one should call from_hist_and_gsr instead of from_gsr")

    @classmethod
    def from_hist_and_gsr_filepaths(cls, hist_filepath: str, gsr_filepath: str,
                                    mng_connector: MongoConnector, with_gsr: bool, with_hist: bool) -> RelaxData:
        """
        Fill the model from a HIST.nc filepath.
        """
        with HistFile(hist_filepath) as hist, GsrFile(gsr_filepath) as gsr:
            return cls.from_hist_and_gsr(hist, gsr, mng_connector, with_gsr, with_hist)

    @classmethod
    def from_hist_and_gsr(cls, hist: HistFile, gsr: GsrFile, mng_connector: MongoConnector,
                          with_gsr: bool, with_hist: bool) -> RelaxData:
        """
        Fill the model from a |HistFile|
        """
        data = ScfData.from_gsr(gsr, mng_connector, with_gsr).dict()

        #an = self.get_relaxation_analyzer()
        #app("Volume change in percentage: %.2f%%" % (an.get_percentage_volume_change() * 100))
        #d = an.get_percentage_lattice_parameter_changes()
        #vals = tuple(d[k] * 100 for k in ("a", "b", "c"))
        #app("Percentage lattice parameter changes:\n\ta: %.2f%%, b: %.2f%%, c: %2.f%%" % vals)
        #an.get_percentage_bond_dist_changes(max_radius=3.0)

        data.update(dict(
            num_steps=hist.num_steps,
            #final_structure_data=StructureData.from_structure(hist.final_structure),
            #etotal_ev_hist=hist.etotals.tolist(),
            #pressure_gpa_hist=,
            #angles_hist=,
            #lenghts_ang_hist=,
            #volume_ang3_hist=,
        ))

        if with_hist:
            data["hist_gfsd"] = mng_connector.gfs_put_filepath(hist.filepath)

        return cls(**data)




# coding: utf-8
"""
Objects to analyze elastic and piezoelectric tensors computed by anaddb.
"""
from __future__ import annotations

import pandas as pd

from collections import OrderedDict
from monty.string import list_strings, marquee
from monty.collections import AttrDict
from monty.json import MSONable
from abipy.core.structure import Structure
from abipy.core.mixins import Has_Structure
from abipy.tools.tensors import Tensor, ElasticTensor, PiezoTensor
from abipy.flowtk.netcdf import ETSF_Reader


class MyElasticTensor(ElasticTensor):

    def _repr_html_(self):
        """Integration with jupyter notebooks."""
        return self.get_voigt_dataframe()._repr_html_()

    def get_voigt_dataframe(self, tol=1e-5):
        """
        Return |pandas-DataFrame| with Voigt indices as colums (C-indexing starting from 0).
        Useful to analyze the converge of individual elements of the tensor(s)
        Elements below tol are set to zero.
        """
        tensor = self.zeroed(tol=tol)
        columns = ["xx", "yy", "zz", "yz", "xz", "xy"]
        #columns = ["1", "2", "3", "4", "5", "6"]
        rows = []
        for row in tensor.voigt:
            rows.append({k: v for k, v in zip(columns, row)})

        df = pd.DataFrame(rows, index=columns, columns=columns)
        df.index.name = "Voigt index"
        return df

    def get_elate_html(self, sysname="Unknown System"):
        """
        Return HTML string with ELATE webpage.
        """
        from abipy.tools.elate.elastic import ELATE
        # Symmetrize the matrix else ELATE complains
        matrix = self.voigt
        matrix = (matrix + matrix.T) / 2.0

        # This is needed because ELATE changes sys.stdout without reverting it before returning
        import sys
        save_stdout = sys.stdout
        try:
            html = ELATE(matrix.tolist(), sysname=sysname)
            return html
        finally:
            sys.stdout = save_stdout


class MyPiezoTensor(PiezoTensor):

    def _repr_html_(self):
        """Integration with jupyter notebooks."""
        return self.get_voigt_dataframe()._repr_html_()

    def get_voigt_dataframe(self, tol=1e-5):
        """
        Return |pandas-DataFrame| with Voigt indices as colums (C-indexing starting from 0).
        Useful to analyze the converge of individual elements of the tensor(s)
        Elements below tol are set to zero.
        """
        tensor = self.zeroed(tol=tol)
        index = ["Px", "Py", "Pz"]
        columns = ["xx", "yy", "zz", "yz", "xz", "xy"]
        #index = ["P1", "P2", "P3"]
        #columns = ["1", "2", "3", "4", "5", "6"]
        rows = []
        for irow, row in enumerate(tensor.voigt):
            rows.append({k: v for k, v in zip(columns, row)})

        df = pd.DataFrame(rows, index=index, columns=columns)
        df.index.name = "Voigt index"
        return df


class ElasticData(Has_Structure, MSONable):
    """
    Container with the different elastic and piezoelectric tensors
    computed by anaddb. Data is stored in pymatgen tensor objects.

    Provides methods to analyze/tabulate data
    See also http://progs.coudert.name/elate/mp?query=mp-2172 for a web interface.
    """

    ALL_ELASTIC_TENSOR_NAMES = (
        "elastic_relaxed",
        "elastic_clamped",
        "elastic_stress_corr",
        "elastic_relaxed_fixed_D",
    )

    ALL_PIEZOELECTRIC_TENSOR_NAMES = (
        "piezo_relaxed",
        "piezo_clamped",
        "d_piezo_relaxed",
        "g_piezo_relaxed",
        "h_piezo_relaxed",
    )

    ALL_TENSOR_NAMES = ALL_ELASTIC_TENSOR_NAMES + ALL_PIEZOELECTRIC_TENSOR_NAMES

    TYPE2NAMES = {
        "all": ALL_TENSOR_NAMES,
        "elastic": ALL_ELASTIC_TENSOR_NAMES,
        "piezoelectric": ALL_PIEZOELECTRIC_TENSOR_NAMES,
    }

    # https://journals.aps.org/prb/abstract/10.1103/PhysRevB.72.035105
    TENSOR_META = {
        "elastic_clamped": AttrDict(
            info="clamped-ion elastic tensor in Voigt notation (shape: (6, 6))",
            units="GPa", latex=r"${\xoverline(C)}$"),
        "elastic_relaxed": AttrDict(
            info="relaxed-ion elastic tensor in Voigt notation (shape: (6, 6))",
            units="GPa", latex=r"${C}$"),
        "elastic_stress_corr": AttrDict(
            info="relaxed-ion elastic tensor considering the stress left inside cell in Voigt notation (shape: (6, 6))",
            units="GPa", latex=r"${C^{\sigma}}$"),
        "elastic_relaxed_fixed_D": AttrDict(
            info="relaxed-ion elastic tensor at fixed displacement field in Voigt notation (shape: (6, 6))",
            units="GPa", latex=r"$C^{(D)}$"),
        "piezo_clamped": AttrDict(
            info="clamped-ion piezoelectric tensor in Voigt notation (shape: (3, 6))",
            units="c/m^2", latex=r"${\xoverline(e)}$"),
        "piezo_relaxed": AttrDict(
            info="relaxed-ion piezoelectric tensor in Voigt notation (shape: (3, 6))",
            units="c/m^2", latex=r"${e}$"),
        "d_piezo_relaxed": AttrDict(
            info="relaxed-ion piezoelectric d tensor in Voigt notation (shape: (3, 6))",
            units="pc/m^2", latex=r"${d}$"),
        "g_piezo_relaxed": AttrDict(
            info="relaxed-ion piezoelectric g tensor in Voigt notation (shape: (3, 6))",
            units="m^2/c", latex=r"${g}$"),
        "h_piezo_relaxed": AttrDict(
            info="relaxed-ion piezoelectric h tensor in Voigt notation (shape: (3, 6))",
            units="GN/c", latex=r"${h}$"),
    }

    def __init__(self, structure, params, elastic_clamped=None, elastic_relaxed=None, elastic_stress_corr=None,
                 elastic_relaxed_fixed_D=None, piezo_clamped=None, piezo_relaxed=None, d_piezo_relaxed=None,
                 g_piezo_relaxed=None, h_piezo_relaxed=None):
        """
        Args:
            structure: |Structure| object.
            params: Dictionary with input parameters.
            elastic_clamped: clamped-ion elastic tensor in Voigt notation in GPa. shape (6,6).
            elastic_relaxed: relaxed-ion elastic tensor in Voigt notation in GPa. shape (6, 6).
            elastic_stress_corr: relaxed-ion elastic tensor considering the stress left inside cell
                in Voigt notation in GPa. shape (6,6).
            elastic_relaxed_fixed_D: relaxed-ion elastic tensor at fixed displacement field
                in Voigt notation in GPa. shape (6,6).
            piezo_clamped: clamped-ion piezoelectric tensor in Voigt notation in c/m^2. shape (3, 6).
            piezo_relaxed: relaxed-ion piezoelectric tensor in Voigt notation in c/m^2. shape (3, 6).
            d_piezo_relaxed: relaxed-ion piezoelectric d tensor in Voigt notation in pc/m^2. shape (3, 6).
            g_piezo_relaxed: relaxed-ion piezoelectric g tensor in Voigt notation in m^2/c. shape (3, 6).
            h_piezo_relaxed: relaxed-ion piezoelectric h tensor in Voigt notation in GN/c. shape (3, 6).

        .. note::

            Arguments can be either arrays or Tensor objects.
        """
        self._structure = structure
        self.params = params
        self.elastic_clamped = self._define_variable(elastic_clamped, MyElasticTensor)
        self.elastic_relaxed = self._define_variable(elastic_relaxed, MyElasticTensor)
        self.elastic_stress_corr = self._define_variable(elastic_stress_corr, MyElasticTensor)
        self.elastic_relaxed_fixed_D = self._define_variable(elastic_relaxed_fixed_D, MyElasticTensor)
        self.piezo_clamped = self._define_variable(piezo_clamped, MyPiezoTensor)
        self.piezo_relaxed = self._define_variable(piezo_relaxed, MyPiezoTensor)
        self.d_piezo_relaxed = self._define_variable(d_piezo_relaxed, MyPiezoTensor)
        self.g_piezo_relaxed = self._define_variable(g_piezo_relaxed, MyPiezoTensor)
        self.h_piezo_relaxed = self._define_variable(h_piezo_relaxed, MyPiezoTensor)

    def _define_variable(self, tensor_voigt, tensor_class):
        """
        Helper function to set values of a variable
        """
        if isinstance(tensor_voigt, Tensor):
            if not isinstance(tensor_voigt, tensor_class):
                raise TypeError("Expecting tensor class `%s`, received class `%s`" % (
                    tensor_class.__name__, tensor_voigt.__class__.__name__))
            return tensor_voigt
        else:
            return tensor_class.from_voigt(tensor_voigt) if tensor_voigt is not None else None

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self._structure

    @classmethod
    def from_file(cls, filepath: str) -> ElasticData:
        """
        Builds the object from an anaddb.nc file
        """
        with ETSF_Reader(filepath) as reader:
            return cls.from_ncreader(reader)

    @classmethod
    def from_ncreader(cls, reader) -> ElasticData:
        """
        Builds the object from a ETSF_Reader
        """
        structure = reader.read_structure()

        params = AttrDict(
            # NB: asr and chneut are always present in the new anaddb.nc file
            # Use -666 to support old formats.
            asr=int(reader.read_value("asr", default=-666)),
            chneut=int(reader.read_value("chneut", default=-666)),
            elaflag=int(reader.read_value("elaflag", default=0)),
            instrflag=int(reader.read_value("instrflag", default=0)),
            piezoflag=int(reader.read_value("piezoflag", default=0)),
            dieflag=int(reader.read_value("dieflag", default=0)),
        )

        ts = AttrDict({n: None for n in cls.ALL_TENSOR_NAMES})

        # [6, 6] symmetric tensors (written by Fortran, produced in ddb_elast)
        ts.elastic_clamped = reader.read_value("elastic_constants_clamped_ion", default=None)
        ts.elastic_relaxed = reader.read_value("elastic_constants_relaxed_ion", default=None)
        if params.elaflag == 5:
            ts.elastic_stress_corr = reader.read_value("elastic_constants_relaxed_ion_stress_corrected")

        # Written in ddb_piezo
        if params.elaflag == 4:
            ts.elastic_relaxed_fixed_D = reader.read_value("elastic_tensor_relaxed_ion_fixed_D", default=None)

        # [3, 6] tensors (written by Fortran, produced in ddb_piezo).
        ts.piezo_clamped = reader.read_value("piezo_clamped_ion", default=None)
        ts.piezo_relaxed = reader.read_value("piezo_relaxed_ion", default=None)

        if params.piezoflag == 4:
            ts.d_piezo_relaxed = reader.read_value("d_tensor_relaxed_ion", default=None)

        # These are [6, 3] tensors written by Fortran (need to transpose).
        if params.piezoflag in (5, 7):
            ts.g_piezo_relaxed = reader.read_value("g_tensor_relaxed_ion", default=None)
            if ts.g_piezo_relaxed is not None:
                ts.g_piezo_relaxed = ts.g_piezo_relaxed.T.copy()

        if params.piezoflag in (6, 7):
            ts.h_piezo_relaxed = reader.read_value("h_tensor_relaxed_ion", default=None)
            if ts.h_piezo_relaxed is not None:
                ts.h_piezo_relaxed = ts.h_piezo_relaxed.T.copy()

        return cls(structure, params, **ts)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String represention with verbosity level `verbose`."""
        lines = []; app = lines.append
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(marquee("Anaddb Variables", mark="="))
        import json
        app(json.dumps(self.params, indent=2, sort_keys=True))

        for tensor_type in ("elastic", "piezoelectric"):
            name_tensor_list = self.name_tensor_list(tensor_type=tensor_type)
            if name_tensor_list:
                app("")
                app(marquee("%s tensors available" % tensor_type, mark="="))
                for name, tensor in name_tensor_list:
                    meta = self.TENSOR_META[name]
                    is_fit = tensor.is_fit_to_structure(self.structure, tol=1e-2)
                    tol = dict(elastic=1e-3, piezoelectric=1e-5)[tensor_type]
                    app("[%s]" % name.upper())
                    app("%s" % meta.info)
                    app("Units: %s, set to zero below: %s, fit_to_structure: %s" % (meta.units, tol, is_fit))
                    app("")
                    if tensor_type == "elastic":
                        app(self.get_elastic_tensor_dataframe(tensor_name=name, tol=tol).to_string())
                    elif tensor_type == "piezoelectric":
                        app(self.get_piezoelectric_tensor_dataframe(tensor_name=name, tol=tol).to_string())
                    app("")

        return "\n".join(lines)

    def get_tensor(self, tensor_name: str, tol=None):
        """
        Return tensor from its name `tensor_name`.
        Set to zero all entries below `tol` if `tol` is not None.
        """
        tensor = getattr(self, tensor_name)
        if tensor is None: return None
        if tol is not None: tensor = tensor.zeroed(tol=tol)
        return tensor

    def name_tensor_list(self, tensor_names=None, tensor_type="all", tol=None):
        """
        List of (name, tensor) tuples. Only tensors stored in the object are returned.

        Args:
            tensor_names: List of tensor names to select. None means all.
            tensor_type: Select tensors by type. Must be in ["all", "elastic", "piezoelectric"].
            tol: Set to zero all all entries below this threshold
        """
        l = []
        if tensor_names is None:
            for name in self.TYPE2NAMES[tensor_type]:
                tensor = self.get_tensor(name, tol=tol)
                if tensor is not None:
                    l.append((name, tensor))
        else:
            for name in list_strings(tensor_names):
                if name not in self.TYPE2NAMES[tensor_type]:
                    raise ValueError("tensor name %s does not belong to type: `%s`" % (name, tensor_type))
                tensor = self.get_tensor(name, tol=tol)
                if tensor is not None:
                    l.append((name, tensor))

        return l

    def fit_to_structure(self, structure=None, symprec=0.1) -> ElasticData:
        """
        Return new ElasticData object with tensors that are invariant with respect to symmetry
        operations corresponding to `structure`.

        Args:
            structure: |Structure| from which to generate symmetry operations
                If None, the internal structure is used.
            symprec (float): symmetry tolerance for the Spacegroup Analyzer
                used to generate the symmetry operations
        """
        structure = self.structure if structure is None else structure
        kwargs = {name: tensor.fit_to_structure(structure, symprec=symprec)
            for name, tensor in self.name_tensor_list()}

        return self.__class__(structure, self.params, **kwargs)

    def convert_to_ieee(self, structure=None, initial_fit=True, refine_rotation=True) -> ElasticData:
        """
        Return new set of tensors in IEEE format according to the 1987 IEEE standards.

        Args:
            structure: A |Structure| structure associated with the
                tensor to be converted to the IEEE standard
                If None, the internal structure is used
            initial_fit (bool): flag to indicate whether initial
                tensor is fit to the symmetry of the structure.
                Defaults to true. Note that if false, inconsistent
                results may be obtained due to symmetrically
                equivalent, but distinct transformations
                being used in different versions of spglib.
            refine_rotation (bool): whether to refine the rotation
                produced by the ieee transform generator, default True
        """
        kwargs = {}
        structure = self.structure if structure is None else structure
        for name, tensor in self.name_tensor_list():
            # TODO: one should pass the ieee stucture.
            kwargs[name] = tensor.convert_to_ieee(structure,
                initial_fit=initial_fit, refine_rotation=refine_rotation)

        return self.__class__(structure, self.params, **kwargs)

    def get_elastic_tensor_dataframe(self, tensor_name="elastic_relaxed", tol=1e-3) -> pd.DataFrame:
        """
        Args:
            tensor_name:
            tol: set to zero all entries below this threshold.
        """
        tensor = self.get_tensor(tensor_name, tol=tol)
        if tensor is None: return pd.DataFrame()
        columns = ["xx", "yy", "zz", "yz", "xz", "xy"]
        #columns = ["1", "2", "3", "4", "5", "6"]
        rows = []
        for row in tensor.voigt:
            rows.append({k: v for k, v in zip(columns, row)})

        return pd.DataFrame(rows, index=columns, columns=columns)

    def get_piezoelectric_tensor_dataframe(self, tensor_name="piezo_relaxed", tol=1e-5) -> pd.DataFrame:
        """
        Args:
            tensor_name:
            tol: set to zero all entries below this threshold.
        """
        tensor = self.get_tensor(tensor_name, tol=tol)
        if tensor is None: return pd.DataFrame()
        index = ["Px", "Py", "Pz"]
        columns = ["xx", "yy", "zz", "yz", "xz", "xy"]
        #index = ["P1", "P2", "P3"]
        #columns = ["1", "2", "3", "4", "5", "6"]
        rows = []
        for row in tensor.voigt:
            rows.append({k: v for k, v in zip(columns, row)})

        return pd.DataFrame(rows, index=index, columns=columns)

    def get_elastic_voigt_dataframe(self, voigt_as_index=True, tol=None):
        return self.get_voigt_dataframe(tensor_names=self.ALL_ELASTIC_TENSOR_NAMES,
                                        voigt_as_index=voigt_as_index, tol=tol)

    def get_piezo_voigt_dataframe(self, voigt_as_index=True, tol=None):
        return self.get_voigt_dataframe(tensor_names=self.ALL_PIEZOELECTRIC_TENSOR_NAMES,
                                        voigt_as_index=voigt_as_index, tol=tol)

    def get_voigt_dataframe(self, tensor_names, voigt_as_index=True, tol=None):
        """
        Return |pandas-DataFrame| with Voigt indices as colums (C-indexing starting from 0).
        Useful to analyze the converge of individual elements of the tensor(s)

        Args:
            tensor_names: List of tensor names.
            tol: Set to zero all all entries below a certain threshold
        """
        rows, index = [], []
        for name, tensor in self.name_tensor_list(tensor_names=tensor_names, tol=tol):
            voigt_map = tensor.get_voigt_dict(tensor.rank)
            row = {}
            for ind in voigt_map:
                row[voigt_map[ind]] = tensor[ind]
            index.append(name)
            row = OrderedDict(sorted(row.items(), key=lambda item: item[0]))
            row["tensor_name"] = name
            rows.append(row)

        # Build dataframe.
        df = pd.DataFrame(rows, index=index, columns=list(rows[0].keys() if rows else None))

        if voigt_as_index:
            # Return transpose to have (i,j) as index and tensor names as columns
            df = df.drop(columns="tensor_name").T
            df.index.name = "voigt_cinds"
            return df.reset_index()
            #return df
        else:
            return df

    def get_elastic_properties_dataframe(self, tensor_names="all", properties_as_index=False,
            include_base_props=True, ignore_errors=False, fit_to_structure=False, symprec=0.1) -> pd.DataFrame:
        """
        Return a |pandas-DataFrame| with properties derived from the elastic tensor
        and the associated structure

        Args:
            tensor_names= ["elastic_relaxed", "elastic_clamped", "elastic_stress_corr", "elastic_relaxed_fixed_D"]
            properties_as_index:
            include_base_props (bool): whether to include base properties, like k_vrh, etc.
            ignore_errors (bool): if set to true, will set problem properties
                that depend on a physical tensor to None, defaults to False
            fit_to_structure (bool): If True, properties are computed with the orginal tensors
                and symmetrized tensors. An additional column `fit_to_structure` is added to the dataframe.
            symprec (float): symmetry tolerance for the Spacegroup Analyzer
                used to generate the symmetry operations if `fit_to_structure`
        """
        tensor_names = self.ALL_ELASTIC_TENSOR_NAMES if tensor_names == "all" else list_strings(tensor_names)
        do_fits = [False] if not fit_to_structure else [True, False]
        rows = []
        for name, tensor in self.name_tensor_list(tensor_names=tensor_names):
            for do_fit in do_fits:
                if do_fit:
                    tensor = tensor.fit_to_structure(self.structure, symprec=symprec)
                d = tensor.get_structure_property_dict(self.structure,
                        include_base_props=include_base_props, ignore_errors=ignore_errors)
                d.pop("structure")
                # Add column telling whether fit has been performed
                if len(do_fits) > 1: d["fit_to_structure"] = do_fit
                d["tensor_name"] = name
                rows.append(d)

        df = pd.DataFrame(rows, columns=list(rows[0].keys() if rows else None))

        if properties_as_index:
            # TODO
            # Return transpose to have (i,j) as index and tensor names as columns
            #columns = df.columns
            df = df.drop(columns="tensor_name").T
            df.index.name = "property"
            #df.columns = columns
            return df.reset_index()
        else:
            return df



# coding: utf-8
"""Scissors operator."""
import os
import numpy as np
import pickle

from collections import OrderedDict
from monty.collections import AttrDict
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt


__all__ = [
    "Scissors",
    "ScissorsBuilder",
]


class ScissorsError(Exception):
    """Base class for the exceptions raised by :class:`Scissors`"""


class Scissors(object):
    """
    This object represents an energy-dependent scissors operator.
    The operator is defined by a list of domains (energy intervals)
    and a list of functions defined in these domains.
    The domains should fulfill the constraints documented in the main constructor.

    .. note::

        eV units are assumed.

    The standard way to create this object is via the methods provided by the factory class :class:`ScissorBuilder`.
    Once the instance has been created, one can correct the band structure by calling the `apply` method.
    """
    Error = ScissorsError

    def __init__(self, func_list, domains, residues, bounds=None):
        """
        Args:
            func_list: List of callable objects. Each function takes an eigenvalue and returns
                the corrected value.
            domains: Domains of each function. List of tuples [(emin1, emax1), (emin2, emax2), ...]
            bounds: Specify how to handle energies that do not fall inside one of the domains.
                At present, only constant boundaries are implemented.
            residues: A list of the residues of the fitting per domain

        .. note::

            #. Domains should not overlap, cover e0mesh, and given in increasing order.

            #. Holes are permitted but the interpolation will raise an exception if the
               eigenvalue falls inside the hole.

            #. Errors contains a list of the fitting errors per domain

        """
        # TODO Add consistency check.
        self.func_list = func_list
        self.domains = np.atleast_2d(domains)
        self.residues = residues
        assert len(self.func_list) == len(self.domains)

        # Treat the out-of-boundary conditions. func_low and func_high are used to handle energies
        # that are below or above the min/max energy given in domains.
        blow, bhigh = "c", "c"
        if bounds is not None:
            blow, bhigh = bounds[0][0], bounds[0][1]

        if blow.lower() == "c":
            try:
                self.func_low = lambda x: float(bounds[0][1])
            except Exception:
                x_low = self.domains[0,0]
                fx_low = func_list[0](x_low)
                self.func_low = lambda x: fx_low
        else:
            raise NotImplementedError("Only constant boundaries are implemented")

        if bhigh.lower() == "c":
            try:
                self.func_high = lambda x: float(bounds[1][1])
            except Exception:
                x_high = self.domains[1, -1]
                fx_high = func_list[-1](x_high)
                self.func_high = lambda x: fx_high
        else:
            raise NotImplementedError("Only constant boundaries are implemented")

        # This counter stores the number of points that are out of bounds.
        self.out_bounds = np.zeros(3, dtype=int)

    def apply(self, eig):
        """Correct the eigenvalue eig (eV units)."""
        # Get the list of domains.
        domains = self.domains

        if eig < domains[0,0]:
            # Eig is below the first point of the first domain.
            # Call func_low
            print("left ", eig, " < ", domains[0,0])
            self.out_bounds[0] += 1
            return self.func_low(eig)

        if eig > domains[-1,1]:
            # Eig is above the last point of the last domain.
            # Call func_high
            print("right ", eig, " > ", domains[-1,1])
            self.out_bounds[1] += 1
            return self.func_high(eig)

        # eig is inside the domains: find the domain
        # and call the corresponding function.
        for idx, dms in enumerate(domains):
            if dms[1] >= eig >= dms[0]:
                return self.func_list[idx](eig)

        self.out_bounds[2] += 1
        raise self.Error("Cannot find location of eigenvalue %s in domains:\n%s" % (eig, domains))


class ScissorsBuilder(object):
    """
    This object facilitates the creation of :class:`Scissors` instances.

    Usage:

        builder = ScissorsBuilder.from_file("out_SIGRES.nc")

        # To plot the QP results as function of the KS energy:
        builder.plot_qpe_vs_e0()

        # To select the domains esplicitly (optional but highly recommended)
        builder.build(domains_spin=[[-10, 6.02], [6.1, 20]])

        # To compare the fitted results with the ab-initio data:
        builder.plot_fit()

        # To plot the corrected bands:
        builder.plot_qpbands(abidata.ref_file("si_nscf_WFK.nc"))
    """

    @classmethod
    def from_file(cls, filepath):
        """
        Generate object from (SIGRES.nc) file. Main entry point for client code.
        """
        from abipy.abilab import abiopen
        with abiopen(filepath) as ncfile:
            return cls(qps_spin=ncfile.qplist_spin, sigres_ebands=ncfile.ebands)

    @classmethod
    def pickle_load(cls, filepath):
        """Load the object from a pickle file."""
        with open(filepath, "rb") as fh:
            d = AttrDict(pickle.load(fh))
            # Costruct the object and compute the scissors.
            new = cls(d.qps_spin, d.sigres_ebands)
            new.build(d.domains_spin, d.bounds_spin)
            return new

    def pickle_dump(self, filepath, protocol=-1):
        """Save the object in Pickle format"""
        assert all(s1 == s2 for s1, s2 in zip(self.domains_spin.keys(), self.bounds_spin.keys()))
        assert all(s1 == s2 for s1, s2 in zip(self.domains_spin.keys(), range(self.nsppol)))

        bounds_spin = None
        if any(v is not None for v in self.bounds_spin.values()):
            bounds_spin = [a.tolist() for a in self.bounds_spin.values()]

        # This trick is needed because we cannot pickle bound methods of the scissors operator.
        d = dict(qps_spin=self._qps_spin,
                 sigres_ebands=self.sigres_ebands,
                 domains_spin=[a for a in self.domains_spin.values()],
                 bounds_spin=bounds_spin)

        with open(filepath, "wb") as fh:
            pickle.dump(d, fh, protocol=protocol)

    def __init__(self, qps_spin, sigres_ebands):
        """
        Args:
            qps_spin: List of :class:`QPlist`, for each spin.
            sigres_ebands: |ElectronBands| obtained from the SIGRES file
        """
        # Sort quasiparticle data by e0.
        self._qps_spin = tuple([qps.sort_by_e0() for qps in qps_spin])

        # Compute the boundaries of the E0 mesh.
        e0min, e0max = np.inf, -np.inf
        for qps in self._qps_spin:
            e0mesh = qps.get_e0mesh()
            e0min = min(e0min, e0mesh[0])
            e0max = max(e0max, e0mesh[-1])

        self._e0min, self._e0max = e0min, e0max

        # The KS bands stored in the sigres file (used to compute automatically the boundaries)
        self.sigres_ebands = sigres_ebands

        # Start with default values for domains.
        self.build()

    @property
    def nsppol(self):
        """Number of spins."""
        return len(self._qps_spin)

    @property
    def e0min(self):
        """Minimum KS energy in eV (takes into account spin)"""
        return self._e0min

    @property
    def e0max(self):
        """Maximum KS energy in eV (takes into account spin)"""
        return self._e0max

    @property
    def scissors_spin(self):
        """Returns a tuple of :class:`Scissors` indexed by the spin value."""
        try:
            return self._scissors_spin
        except AttributeError:
            raise AttributeError("Call self.build to create the scissors operator")

    def build(self, domains_spin=None, bounds_spin=None, k=3):
        """
        Build the scissors operator.

        Args:
            domains_spin: list of domains in eV for each spin. If domains is None,
                domains are computed automatically from the sigres bands
                (two domains separated by the middle of the gap).
            bounds_spin: Options specifying the boundary conditions (not used at present)
            k: Parameter defining the order of the fit.
        """
        nsppol = self.nsppol

        # The parameters defining the scissors operator
        self.domains_spin = OrderedDict()
        self.bounds_spin = OrderedDict()

        if domains_spin is None:
            # Use sigres_ebands and the position of the homo, lumo to compute the domains.
            domains_spin = nsppol * [None]
            e_bands = self.sigres_ebands
            for spin in e_bands.spins:
                gap_mid = (e_bands.homos[spin].eig + e_bands.lumos[spin].eig) / 2
                domains_spin[spin] = [[self.e0min - 0.2 * abs(self.e0min), gap_mid],
                                      [gap_mid, self.e0max + 0.2 * abs(self.e0max)]]
                #print("domains", domains_spin[spin])
        else:
            if nsppol == 1:
                domains_spin = np.reshape(domains_spin, (1, -1, 2))
            elif nsppol == 2:
                assert len(domains_spin) == nsppol
                if bounds_spin is not None: assert len(bounds_spin) == nsppol
            else:
                raise ValueError("Wrong number of spins %d" % nsppol)
            #if len(domains_spin) != nsppol:
            #    raise ValueError("len(domains_spin) == %s != nsppol %s" % (len(domains_spin), nsppol))

        # Construct the scissors operator for each spin.
        scissors_spin = nsppol * [None]
        for spin, qps in enumerate(self._qps_spin):
            bounds = None if not bounds_spin else bounds_spin[spin]
            scissors_spin[spin] = qps.build_scissors(domains_spin[spin], bounds=bounds, k=k, plot=False)

            # Save input so that we can reconstruct Scissors.
            self.domains_spin[spin] = domains_spin[spin]
            self.bounds_spin[spin] = bounds

        self._scissors_spin = scissors_spin
        return domains_spin

    @add_fig_kwargs
    def plot_qpe_vs_e0(self, with_fields="all", **kwargs):
        """Plot the quasiparticle corrections as function of the KS energy."""
        ax_list = None
        for spin, qps in enumerate(self._qps_spin):
            kwargs["title"] = "spin %s" % spin
            fig = qps.plot_qps_vs_e0(with_fields=with_fields, ax_list=ax_list, show=False, **kwargs)
            ax_list = fig.axes

        return fig

    @add_fig_kwargs
    def plot_fit(self, ax=None, fontsize=8, **kwargs):
        """
        Compare fit functions with input quasi-particle corrections.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for titles and legend.

        Return: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        for spin in range(self.nsppol):
            qps = self._qps_spin[spin]
            e0mesh, qpcorrs = qps.get_e0mesh(), qps.get_qpeme0().real

            ax.scatter(e0mesh, qpcorrs, label="Input QP corrections, spin %s" % spin)
            scissors = self._scissors_spin[spin]
            intp_qpc = [scissors.apply(e0) for e0 in e0mesh]
            ax.plot(e0mesh, intp_qpc, label="Scissors operator, spin %s" % spin)

        ax.grid(True)
        ax.set_xlabel('KS energy (eV)')
        ax.set_ylabel('QP-KS (eV)')
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    def plot_qpbands(self, bands_filepath, bands_label=None, dos_filepath=None, dos_args=None, **kwargs):
        """
        Correct the energies found in the netcdf file bands_filepath and plot the band energies (both the initial
        and the corrected ones) with matplotlib. The plot contains the KS and the QP DOS if dos_filepath is not None.

        Args:
            bands_filepath: Path to the netcdf file containing the initial KS energies to be corrected.
            bands_label String used to label the KS bands in the plot.
            dos_filepath: Optional path to a netcdf file with the initial KS energies on a homogeneous k-mesh
                (used to compute the KS and the QP dos)
            dos_args: Dictionary with the arguments passed to get_dos to compute the DOS
                Used if dos_filepath is not None.

            kwargs: Options passed to the plotter.

        Return: |matplotlib-Figure|
        """
        from abipy.abilab import abiopen, ElectronBandsPlotter

        # Read the KS band energies from bands_filepath and apply the scissors operator.
        with abiopen(bands_filepath) as ncfile:
            ks_bands = ncfile.ebands
            #structure = ncfile.structure

        qp_bands = ks_bands.apply_scissors(self._scissors_spin)

        # Read the band energies computed on the Monkhorst-Pack (MP) mesh and compute the DOS.
        ks_dos, qp_dos = None, None
        if dos_filepath is not None:
            with abiopen(dos_filepath) as ncfile:
                ks_mpbands = ncfile.ebands

            dos_args = {} if not dos_args else dos_args
            ks_dos = ks_mpbands.get_edos(**dos_args)
            # Compute the DOS with the modified QPState energies.
            qp_mpbands = ks_mpbands.apply_scissors(self._scissors_spin)
            qp_dos = qp_mpbands.get_edos(**dos_args)

        # Plot the LDA and the QPState band structure with matplotlib.
        plotter = ElectronBandsPlotter()

        bands_label = bands_label if bands_label is not None else os.path.basename(bands_filepath)
        plotter.add_ebands(bands_label, ks_bands, edos=ks_dos)
        plotter.add_ebands(bands_label + " + scissors", qp_bands, edos=qp_dos)

        #qp_marker: if int > 0, markers for the ab-initio QP energies are displayed. e.g qp_marker=50
        #qp_marker = 50
        #if qp_marker is not None:
        #    # Compute correspondence between the k-points in qp_list and the k-path in qp_bands.
        #    # TODO
        #    # WARNING: strictly speaking one should check if qp_kpoint is in the star of k-point.
        #    # but compute_star is too slow if written in pure python.
        #    x, y, s = [], [], []
        #    for ik_path, kpoint in enumerate(qp_bands.kpoints):
        #        #kstar = kpoint.compute_star(structure.fm_symmops)
        #        for spin in range(self.nsppol):
        #            for ik_qp, qp in enumerate(self._qps_spin[spin]):
        #                #if qp.kpoint in kstar:
        #                if qp.kpoint == kpoint:
        #                    x.append(ik_path)
        #                    y.append(np.real(qp.qpe))
        #                    s.append(qp_marker)
        #    plotter.set_marker("ab-initio QP", [x, y, s])

        return plotter.combiplot(**kwargs)



#!/usr/bin/env python
r"""
K-path from IBZ
===============

This example shows how to extract energies along a k-path
from a calculation done with a (dense) IBZ sampling.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the file with energies computed with a homogeneous sampling of the BZ
# and extract the band structure object.
with abiopen(abidata.ref_file("si_scf_GSR.nc")) as gs_file:
    ebands_ibz = gs_file.ebands

# This is a GS calculation done with a 8x8x8 k-mesh.
print(ebands_ibz.kpoints)

# Build new ebands with energies along G-X-L-G path.
# Smooth bands require dense k-meshes.
r = ebands_ibz.with_points_along_path(knames=["G", "X", "L", "G"])

print(r.ebands)
r.ebands.plot()



""""MongoGUI."""
from __future__ import annotations

import json
#import param
import functools
import panel as pn
import panel.widgets as pnw

from typing import Type
from abipy.panels.core import AbipyParameterized, Loading  # depends_on_btn_click, ply,
from abipy.htc.base_models import QueryResults, MongoConnector
from abipy.htc.flow_models import FlowModel


def display_qr(method):
    @functools.wraps(method)
    def decorated(*args, **kwargs):
        self = args[0]
        with Loading(self.out_area):
            qr = method(*args, **kwargs)
            self._display_qr(qr)
    return decorated


class MongoGui(AbipyParameterized):
    """
    A MongoGUI is connected to MongoDB collection FlowModel documents (actually, a concrete
    subclass of FlowModel) via a MongoConnector.
    The GUI uses the abstract protocol of the FlowModel to provide widgets to filter
    analyze, aggregate and visualize the results.
    The abstract protocol should be already enough to cover the most important cases.
    In needed, one can define a subclass of MongoGUI providing extra tools to interact with that
    particular FlowModel.
    """

    def __init__(self, mng_connector: MongoConnector, flow_model_cls: Type[FlowModel]):
        super().__init__()
        self.mng_connector = mng_connector
        self.flow_model_cls = flow_model_cls

        self.build_ui()

    #def collection(self):
    #    return self.mng_connector.get_collection()

    def build_ui(self):
        # This is the part that should be abstracted out
        # so that one can implement customized subclasses
        # that rely on model.get_panel_view

        self.limit = 0

        self.filter_query = pnw.LiteralInput(name='MongoDB filter (python dict)',
                                             placeholder="e.g. {'key': value}", type=dict,
                                             value={'kppa': 300})

        self.filter_btn = pnw.Button(name="Run query", button_type='primary')
        self.filter_btn.on_click(self.on_filter_btn)

        collection = self.mng_connector.get_collection()

        # Get all spg_numbers in the collection.
        spg_numbers = self.flow_model_cls.mng_get_spg_numbers_incoll(collection)
        self.spg_number_wdg = pnw.Select(name='Filter by space group number', options=spg_numbers)
        #self.spg_number_wdg = pnw.IntInput(name='Filter by space group number', value=None, step=1, start=1, end=230)

        self.filter_by_spg_number_btn = pnw.Button(name="Run query", button_type='primary')
        self.filter_by_spg_number_btn.on_click(self.on_filter_by_spg_number_btn)

        self.formula_wdg = pnw.TextInput(name='Filter by reduced formula', placeholder='Enter e.g. Si or MgO...')
        self.filter_by_formula_btn = pnw.Button(name="Run query", button_type='primary')
        self.filter_by_formula_btn.on_click(self.on_filter_by_formula_btn)

        crystal_systems = self.flow_model_cls.mng_get_crystal_systems_incoll(collection)
        self.crystal_system_wdg = pnw.Select(name='Filter by crystal system', options=crystal_systems)
        self.filter_by_crystal_system_btn = pnw.Button(name="Run query", button_type='primary')
        self.filter_by_crystal_system_btn.on_click(self.on_filter_by_crystal_system_btn)

        self.preset_queries_btn = pnw.Button(name="Run query", button_type='primary')
        self.preset_queries_btn.on_click(self.on_preset_queries_btn)
        options = []
        # TODO: Fixme
        #if hasattr(self.flow_model_cls, "get_preset_queries"):
        #    options = [json.dumps(q) for q in self.flow_model_cls.get_preset_queries()]
        self.preset_queries_wdg = pnw.Select(name="Pommon MongoDB queries", options=options)

        md = pn.pane.Markdown("## Perform MongoDB queries using one of the widgets below:")

        widget_box = pn.WidgetBox(
            md,
            self.filter_query, self.filter_btn,
            pn.Row(self.spg_number_wdg, self.formula_wdg),
            pn.Row(self.filter_by_spg_number_btn, self.filter_by_formula_btn),
            self.crystal_system_wdg, self.filter_by_crystal_system_btn,
            self.preset_queries_wdg, self.preset_queries_btn,
            #sizing_mode="stretch_width",
        )

        self.out_area = pn.Column(sizing_mode="stretch_width")

        self.main = pn.Column(self.mng_connector,
                              widget_box,
                              #pn.layout.Divider(),
                              #"## Results:",
                              self.out_area,
                              sizing_mode="stretch_width")

    def get_app(self):
        """Build and return the application."""
        template = pn.template.FastListTemplate
        #template = pn.template.BootstrapTemplate
        #template = pn.template.ReactTemplate
        return template(main=self.main, title="AbiPyFlowModel MongoDB GUI")

    #@depends_on_btn_click('filter_btn', show_shared_wdg_warning=False)
    def on_filter_btn(self, _):
        with Loading(self.out_area):
            collection = self.mng_connector.get_collection()
            query = self.filter_query.value
            qr = self.flow_model_cls.mng_find(query, collection, limit=self.limit)
            self._display_qr(qr)

    @display_qr
    def on_filter_by_spg_number_btn(self, _):
        """Filter documents by space group number."""
        spg_number = self.spg_number_wdg.value
        collection = self.mng_connector.get_collection()
        qr = self.flow_model_cls.mng_find_by_spg_number(spg_number, collection, limit=self.limit)
        return qr

    @display_qr
    def on_filter_by_formula_btn(self, _):
        """Filter documents by formula."""
        reduced_formula = self.formula_wdg.value
        collection = self.mng_connector.get_collection()
        qr = self.flow_model_cls.mng_find_by_formula(reduced_formula, collection, limit=self.limit)
        return qr

    @display_qr
    def on_filter_by_crystal_system_btn(self, _):
        """Filter documents by crystal system."""
        crystal_system = self.crystal_system_wdg.value
        collection = self.mng_connector.get_collection()
        qr = self.flow_model_cls.mng_find_by_crystal_system(crystal_system, collection, limit=self.limit)
        return qr

    @display_qr
    def on_preset_queries_btn(self, _):
        """
        Show the list of preset queries if not index is provided by the user
        else perform the query and return the results.
        """
        query = self.preset_queries_wdg.value
        collection = self.mng_connector.get_collection()
        if query is None:
            return QueryResults.empty_from_query({})

        query = json.loads(query)
        qr = self.flow_model_cls.mng_find(query, collection, limit=self.limit)
        return qr

    def _display_qr(self, qr):
        """
        Helper function that builds the view and update `self.out_area`.
        """
        objects = []
        if qr:
            for model in qr.models:
                objects.append(model.get_panel_view(self.mng_connector))
        else:
            alert = pn.pane.Alert(f"No model found in collection for query: {qr.query}.",
                                  alert_type="danger")
            objects.append(alert)

        self.out_area.objects = objects



"""
Tools and helper functions to build the command line interface of the AbiPy scripts.
"""
from __future__ import annotations

import argparse
import os

from pprint import pformat


def user_wants_to_abort() -> bool:
    """Interactive prompt, return False if user entered `n` or `no`."""
    try:
        answer = input("\nDo you want to continue [Y/n]")
    except EOFError:
        return False

    return answer.lower().strip() in ["n", "no"]


def set_loglevel(loglevel: str) -> None:
    # loglevel is bound to the string value obtained from the command line argument.
    # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
    import logging
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)
    logging.basicConfig(level=numeric_level)


def pn_serve_parser(**kwargs) -> argparse.ArgumentParser:
    """
    Parent parser implementing cli options for panel.serve
    """
    p = argparse.ArgumentParser(add_help=False)

    p.add_argument("--port", default=0, type=int, help="Port to listen on.")
    p.add_argument("--address", default=None,
                              help="The address the server should listen on for HTTP requests.")
    #p.add_argument("--show", default=True, action="store_true", help="Open app in web browser")
    p.add_argument("--num_procs", default=1, type=int,
                              help="Number of worker processes for the app. Defaults to 1")
    p.add_argument('--panel-template', "-pnt", default="FastList",
                  help="Specify template for panel dasboard." +
                       "Possible values are: FastList, FastGrid, Golden, Bootstrap, Material, React, Vanilla." +
                       "Default: FastList")
    p.add_argument('--has-remote-server', default=False, action="store_true",
                  help="True if we are running on the ABINIT server. " +
                       "This flag activates limitations on what the user can do." +
                       "Default: False")
    p.add_argument("--websocket-origin", default=None, type=str,
            help="Public hostnames which may connect to the Bokeh websocket.\n Syntax: " +
                  "HOST[:PORT] or *. Default: None")
    p.add_argument('--max_size_mb', default=150, type=int,
                help="Maximum message size in Mb allowed by Bokeh and Tornado. Default: 150")

    p.add_argument('--no-browser', action='store_true', default=False,
                   help=("Start the jupyter server to serve the notebook "
                         "but don't open the notebook in the browser.\n"
                         "Use this option to connect remotely from localhost to the machine running the kernel"))

    return p


def get_pn_serve_kwargs(options) -> dict:
    """
    Return dict with the arguments to be passed to pn.serve.
    """
    import abipy.panels as mod
    assets_path = os.path.join(os.path.dirname(mod.__file__), "assets")

    serve_kwargs = dict(
        address=options.address,
        port=options.port,
        #dev=True,
        #start=True,
        #show=options.show,
        show=not options.no_browser,
        debug=options.verbose > 0,
        #title=app_title,
        num_procs=options.num_procs,
        static_dirs={"/assets": assets_path},
        websocket_origin=options.websocket_origin,
        #
        # Increase the maximum websocket message size allowed by Bokeh
        # https://panel.holoviz.org/reference/widgets/FileInput.html
        websocket_max_message_size=options.max_size_mb * 1024**2,
        # Increase the maximum buffer size allowed by Tornado
        http_server_kwargs={'max_buffer_size': options.max_size_mb * 1024**2},
    )

    if getattr(options, "verbose"):
        print("Calling pn.serve with serve_kwargs:\n", pformat(serve_kwargs), "\n")

    if options.no_browser:
        print("""
Use:

    ssh -N -f -L localhost:{port}:localhost:{port} username@your_remote_cluster

for port forwarding.
""")


    return serve_kwargs


def customize_mpl(options) -> None:

    if options.mpl_backend is not None:
        # Set matplotlib backend
        import matplotlib
        matplotlib.use(options.mpl_backend)

    if options.seaborn:
        # Use seaborn settings.
        import seaborn as sns
        sns.set(context=options.seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)


def add_expose_options_to_parser(parser, with_mpl_options=True) -> None:
    """
    Add Expose options to the parser.
    """

    parser.add_argument('-e', '--expose', action='store_true', default=False,
        help="Open file and generate matplotlib figures automatically by calling expose method.")
    parser.add_argument("-s", "--slide-mode", default=False, action="store_true",
        help="Iterate over figures. Expose all figures at once if not given on the CLI.")
    parser.add_argument("-t", "--slide-timeout", type=int, default=None,
        help="Close figure after slide-timeout seconds (only if slide-mode). Block if not specified.")
    parser.add_argument("-ew", "--expose-web", default=False, action="store_true",
            help='Generate matplotlib plots in $BROWSER instead of X-server. WARNING: Not all the features are supported.')
    parser.add_argument("-ply", "--plotly", default=False, action="store_true",
            help='Generate plotly plots in $BROWSER instead of matplotlib. WARNING: Not all the features are supported.')
    parser.add_argument("-cs", "--chart-studio", default=False, action="store_true",
            help="Push figure to plotly chart studio ." +
                 "Requires --plotly option and user account at https://chart-studio.plotly.com.")

    if with_mpl_options:
        parser.add_argument('-sns', "--seaborn", const="paper", default=None, action='store', nargs='?', type=str,
            help='Use seaborn settings. Accept value defining context in ("paper", "notebook", "talk", "poster"). Default: paper')
        parser.add_argument('-mpl', "--mpl-backend", default=None,
            help=("Set matplotlib interactive backend. "
                  "Possible values: GTKAgg, GTK3Agg, GTK, GTKCairo, GTK3Cairo, WXAgg, WX, TkAgg, Qt4Agg, Qt5Agg, macosx."
                  "See also: https://matplotlib.org/faq/usage_faq.html#what-is-a-backend."))


class EnumAction(argparse.Action):
    """
    Argparse action for handling Enums

    Usage:

        class Do(enum.Enum):
            Foo = "foo"
            Bar = "bar"

        parser = argparse.ArgumentParser()
        parser.add_argument('do', type=Do, action=EnumAction)

    Taken from https://stackoverflow.com/questions/43968006/support-for-enum-arguments-in-argparse
    """
    def __init__(self, **kwargs):
        # Pop off the type value
        enum_type = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum_type is None:
            raise ValueError("type must be assigned an Enum when using EnumAction")
        if not issubclass(enum_type, enum.Enum):
            raise TypeError("type must be an Enum when using EnumAction")

        # Generate choices from the Enum
        kwargs.setdefault("choices", tuple(e.value for e in enum_type))

        super().__init__(**kwargs)

        self._enum = enum_type

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        value = self._enum(values)
        setattr(namespace, self.dest, value)


def fix_omp_num_threads() -> int:
    """
    Set OMP_NUM_THREADS to 1 if env var is not defined. Return num_threads.
    """
    num_threads = os.getenv("OMP_NUM_THREADS", default=None)
    if num_threads is None:
        num_threads = 1
        os.environ["OMP_NUM_THREADS"] = str(num_threads)

    return num_threads


def range_from_str(string: str) -> range:
    """
    Convert string into a range object.
    """
    if string is None: return None

    tokens = string.split(":")
    start, stop, step = 0, None, 1
    if len(tokens) == 1:
        stop = int(tokens[0])
    elif len(tokens) == 2:
        start, stop = map(int, tokens)
    elif len(tokens) == 3:
        start, stop, step = map(int, tokens)
    else:
        raise ValueError(f"Cannot interpret {string=} as range object.")

    return range(start, stop, step)




# coding: utf-8
"""
Work subclasses for the computation Linear Response Hubbard U and Hund's J with LRUJ post-processor
"""
from __future__ import annotations

from abipy.core.structure import Structure
from abipy.flowtk.works import Work

from typing import List, Any


class LrujWork(Work):

    @classmethod
    def from_pawujv(cls, scf_input, pawujv_ha_list, workdir=None, manager=None) -> LrujWork:
        work = cls(workdir=workdir, manager=manager)

        #scf_task = work.register_scf_task(scf_input)
        #for pawujv in pawujv_ha:
        #    new_inp = scf_input.new_with_vars(pawujv=pawujv, prtwf=-1)
        #    work.register_scf_task(new_inp, deps={scf_task: "WFK"})

        return work

    def on_all_ok(self):
        """
        This method is called once the `Work` is completed i.e. when all tasks have reached status S_OK.
        """



def test_lruj_work():
    # Antiferro. II NiO, 4 atom | Hubbard U on Ni 3d | LMac 2022  ##
    from abipy.core.structure import Structure

    pseudos = []
    structure = Structure.from_abistring("""
#################################################################
## Automatic test/tutorial for ABINIT:                         ##
## Linear Response Hubbard U and Hund's J (LRUJ)               ##
## post-processing utility for the determination of            ##
## said parameters in situ and ab initio                       ##
##    Initialization of WFK files                              ##
##                                                             ##
## Antiferro. II NiO, 4 atom | Hubbard U on Ni 3d | LMac 2022  ##
#################################################################

#Run Parameters
nstep 30                          #Higher than normal because of magnetic state
tolvrs 10d-12
ecut 10
pawecutdg 20
chkprim 0                         #Will complain otherwise with AFII magnetic state
occopt 0
nband 26                          #24 occupied + 2 unoccupied
occ  24*1.0 2*0.0 24*1.0 2*0.0

#Structural Parameters
natom 4                           #NOTE: Converge U(J) wrt supercell size!
ntypat 3                          #Specify perturbed Ni atom as a separate species
typat 1 2 3 3                     #to that of the other Ni atom.
znucl 28 28 8                     #First two are Ni atoms, last two are O
acell 3*7.8800
xred  0.0000000000 0.0000000000 0.0000000000
      0.5000000000 0.5000000000 0.5000000000
      0.2500000000 0.2500000000 0.2500000000
      0.7500000000 0.7500000000 0.7500000000

rprim 0.5000000000 0.5000000000 1.0000000000
      1.0000000000 0.5000000000 0.5000000000
      0.5000000000 1.0000000000 0.5000000000

#Spin Parameters
nsppol 2
nspden 2
nspinor 1
spinat 0 0 3                      #Set high to enforce convergence to high magnetic state
       0 0 -3                     #Otherwise, it may collapse to non-magnetic state
       0 0 0
       0 0 0

#Kpoint Grid
kptopt 1                          #Monkhorst-Pack Mesh
chksymbreak 0                     #Don't check for symmetry breaking
ngkpt 4 4 4

# DFT+U
usepawu 1                         #Alert Abinit to use of DFT+U
lpawu 2 2 1                       #Subspaces of interest: Ni 3d, O 2p
upawu 0.0 0.0 0.0 eV              #Raw (non-corrected) XC functional required to establish U(J)
jpawu 0.0 0.0 0.0 eV
dmatpuopt 3

#Pseudos
pp_dirpath "$ABI_PSPDIR/Pseudodojo_paw_pbe_standard/"
pseudos "Ni.xml,Ni.xml,O.xml"     #Use same pseudopotential for both Ni atoms
""")
    print(structure)
    print(structure.to_abivars())

    scf_input = None

    pawujv_ha = [1, 2, 3]
    work = LrujWork.from_pawujv(scf_input, pawujv_ha)
    print(work)


if __name__ == "__main__":
    test_lruj_work()



from __future__ import annotations

#from typing import List  # , Tuple, ClassVar, Union, TypeVar, Type, Dict  #, Optional, Any, Type,
from pydantic import Field, validator
from abipy.abio.inputs import AbinitInput
from abipy.flowtk import TaskManager, PhononFlow
from .base_models import MongoConnector, GfsFileDesc
from .worker import AbipyWorker
from .gs_models import ScfData
from .dfpt_models import PhononData
from .flow_models import FlowModel, PresetQuery


class _BasePhononFlowModel(FlowModel):
    """
    This is the base class for models performing phonon band structure calculations.

    It defines ...
    as well as submodels storing the output results.
    """

    ########
    # Input
    ########

    with_becs: bool = Field(..., description="Activate computation of Born effective charges.")

    with_quad: bool = Field(..., description="Activate calculation of dynamical quadrupoles.")

    with_flexoe: bool = Field(False, description="Activate computation of the flexoelectric tensor.")

    with_gsden: bool = Field(True, description="Set it to True to save the DEN file in GridFS.")

    with_gspot: bool = Field(True, description="Set it to True to save the POT file in GridFS.")

    with_dvdb: bool = Field(True, description="False if the DVDB file should not be added to GridFs")

    ########################################
    # Input/Output depending on the subclass
    ########################################

    scf_input: AbinitInput = Field(..., description="Input for the initial SCF run")

    ########
    # Output
    ########

    scf_data: ScfData = Field(None, description="Results produced by the initial GS SCF run.")

    phonon_data: PhononData = Field(None, description="Results produced by the Phonon calculation")

    gsden_gfsd: GfsFileDesc = Field(None, description="Pointer to the GS DEN file in GridFS.")

    gspot_gfsd: GfsFileDesc = Field(None, description="Pointer to the GS KS POT file in GridFS.")

    ddb_gfsd: GfsFileDesc = Field(None, description="Pointer to the DDB file from GridFS.")

    dvdb_gfsd: GfsFileDesc = Field(None, description="Pointer to the DVDB Fortran file in GridFS.")

    def build_flow(self, workdir: str, worker: AbipyWorker) -> PhononFlow:
        """
        Build an AbiPy Flow in workdir using the input data available in the model and return it.
        """
        # TODO: Add option to compute bands, get_dvdb filepath!
        flow = PhononFlow.from_scf_input(workdir, self.scf_input,
                                         ph_ngqpt=(2, 2, 2),
                                         #ph_ngqpt=(4, 4, 4),
                                         with_becs=self.with_becs,
                                         with_quad=self.with_quad,
                                         with_flexoe=self.with_flexoe,
                                         )

        # Make sure Abinit writes the output files required by the model
        scf_input = flow[0][0].input
        if self.with_gsden: scf_input.set_vars(prtden=1)
        if self.with_gspot: scf_input.set_vars(prtpot=1)

        return flow

    def postprocess_flow(self, flow: PhononFlow, worker: AbipyWorker) -> None:
        """
        Analyze the flow and fills the model with output results.
        MongoConnector should be used only to insert files in GridFs as the final insertion is done by the caller.
        This function is called by the worker if the flow completed successfully.
        """
        mng_connector = worker.mng_connector

        scf_task = flow[0][0]
        with scf_task.open_gsr() as gsr:
            self.scf_data = ScfData.from_gsr(gsr, mng_connector, with_gsr=False)

        # Add DEN and POT files to GridFs.
        if self.with_gsden:
            self.gsden_gfsd = mng_connector.gfs_put_filepath(scf_task.outdir.need_abiext("DEN"))

        if self.with_gspot:
            self.gspot_gfsd = mng_connector.gfs_put_filepath(scf_task.outdir.need_abiext("POT"))

        with flow.open_final_ddb() as ddb:
            self.phonon_data = PhononData.from_ddb(ddb, mng_connector)
            self.ddb_gfsd = mng_connector.gfs_put_filepath(ddb.filepath)

        if self.with_dvdb:
            # Add to GridDF the DVDB required for e-ph computations.
            ph_work = flow[1]
            dvdb_filepath = ph_work.outdir.need_abiext("DVDB")
            self.dvdb_gfsd = mng_connector.gfs_put_filepath(dvdb_filepath)

    @classmethod
    def get_preset_queries(cls) -> list[PresetQuery]:
        """
        Return list of dictionaries with the MongoDB queries typically used to filter documents for this model.
        Empty list if no suggestion is available.
        """
        return [
            PresetQuery.for_large_forces_or_high_pressure("scf_data", cls),
            #PresetQuery.unstable_ph_modes("phonon_data", cls),
        ]

    def get_panel_view(self, mng_connector: MongoConnector):
        """
        Return panel object with a view of the model.
        """
        raise NotImplementedError()


class PhononFlowModelWithParams(_BasePhononFlowModel):
    """
    This model generates the SCF input from meta parameters that will be stored in the database.
    """

    #kppa: float
    #qppa: float

    def __init__(self, **data):
        super().__init__(**data)

        self.make_input_from_params()
        if self.scf_input is None:
            raise ValueError("scf_input should be defined after the call to `make_input_from_params")


    def make_input_from_params(self):
        if self.scf_input is not None:
            raise ValueError(f"scf_input should be None when calling {self.__class__.__name__} "
                             "as scf_input is automatically generated from the meta-parameters.")
        pseudos = self.pseudos_specs.get_pseudos()
        structure = self.in_structure_data.structure
        # Call the factory function to build the scf_input from the meta-parameters.
        #self.scf_input =


class PhononFlowModelWithInput(_BasePhononFlowModel):
    """
    This model requires a scf AbinitInput constructed by the user.
    It is more flexible than the parameter-based version but queries related to input variables
    are more complex to perform as one should use the |AbinitInput| dictionary.
    """

    #ph_ngqpt: Tuple[int, int, int]

    @validator("scf_input")
    def check_scf_input_is_not_none(cls, v):
        if v is None:
            raise ValueError(f"Model {cls.__name__} requires an scf_input object.")
        return v

    #@root_validator
    #def check_inputs(cls, values):
    #    scf_input, nscf_input = values.get('scf_input'), values.get('nscf_input')
    #    if scf_input is None or nscf_input is None:
    #        raise ValueError(f"Constructing a {cls.__name__} model requires both scf_input and nscf_input")
    #    return values



#!/usr/bin/env python
from __future__ import print_function, unicode_literals, division, absolute_import
import sys

from abipy.data import AnaddbFilesGenerator


class MyGenerator(AnaddbFilesGenerator):
    """This class generates the output files used in the unit tests and in the examples."""

    def __init__(self, **kwargs):
        super(MyGenerator, self).__init__(**kwargs)

        self.files_to_keep.add("ddb_notes")


    # Mapping old_name --> new_name for the output files that must be preserved.
    files_to_save = {
        "out_PHBST.nc": "ZnSe_hex_886.out_PHBST.nc",
        "out_PHDOS.nc": "ZnSe_hex_886.out_PHDOS.nc",
        "anaddb.nc": "ZnSe_hex_886.anaddb.nc",
    }

    in_ddb = "ZnSe_hex_qpt_DDB"


if __name__ == "__main__":
    sys.exit(MyGenerator().run())



#!/usr/bin/env python
r"""
Band structure interpolation with Wannier functions
===================================================

This example shows how to analyze the wannier90 results using
the ABIWAN.nc netcdf file produced by Abinit when calling wannier90 in library mode.
Use `abiopen FILE.wout` for a command line interface and
the `--expose` option to generate matplotlib figures automatically.
"""
import os
import abipy.data as abidata

from abipy.abilab import abiopen

# Open the ABIWAN file
filepath = os.path.join(abidata.dirpath, "refs", "wannier90", "tutoplugs_tw90_4", "tw90_4o_DS3_ABIWAN.nc")
abiwan = abiopen(filepath)
print(abiwan)

# Plot the matrix elements of the KS Hamiltonian in real space in the Wannier Gauge.
abiwan.hwan.plot(title="Matrix elements in real space")

# Interpolate bands with Wannier functions using an automatically selected k-path
# with 5 points for the smallest segment.
ebands_wan_kpath = abiwan.interpolate_ebands(line_density=5)
ebands_wan_kpath.plot(title="Wannier-interpolated")

# Interpolate bands in the IBZ defined by ngkpt
ebands_wan_kmesh = abiwan.interpolate_ebands(ngkpt=[8, 8, 8])
edos = ebands_wan_kmesh.get_edos()
ebands_wan_kpath.plot_with_edos(edos, title="Wannier-interpolated bands and DOS")

# To compare the interpolated bands with ab-initio results,
# pass a file with the ab-initio bands to the get_plotter_from_ebands method
# that will return an ElectronBandsPlotter object.

#plotter = abiwan.get_plotter_from_ebands("out_GSR.nc")
#plotter.combiplot()



#!/usr/bin/env python
"""
GW corrections
==============

This example shows how to interpolate the GW corrections and use the interpolated
values to correct the KS band structure computed on a high symmetry k-path and
the KS energies of a k-mesh. Finally, the KS and the GW results are plotted with matplotlib.
"""
import abipy.data as abidata
from abipy.abilab import abiopen, ElectronBandsPlotter

with abiopen(abidata.ref_file("si_scf_GSR.nc")) as gsr:
    ebands_ibz = gsr.ebands

with abiopen(abidata.ref_file("si_nscf_GSR.nc")) as gsr:
    ebands_kpath = gsr.ebands

skw = ebands_kpath.interpolate(lpratio=5, filter_params=None)

skw.ebands_kpath.plot()

#plotter = ebands_kpath.get_plotter_with("Ab-initio", "SKW interp", skw.ebands_kpath)
#plotter.combiplot()

"""
# Get quasiparticle results from the SIGRES.nc database.
sigres = abiopen(abidata.ref_file("si_g0w0ppm_nband30_SIGRES.nc"))

# Read the KS band energies computed on the k-path
with abiopen(abidata.ref_file("si_nscf_GSR.nc")) as gsr_nscf:
    ks_ebands_kpath = gsr_nscf.ebands

# Read the KS band energies computed on the Monkhorst-Pack (MP) mesh
# and compute the DOS with the Gaussian method
with abiopen(abidata.ref_file("si_scf_GSR.nc")) as gsr_scf:
    ks_ebands_kmesh = gsr_scf.ebands

ks_edos = ks_ebands_kmesh.get_edos()

# Interpolate the QP corrections and use the interpolated values to correct
# the KS energies stored in `ks_ebands_kpath` and `ks_ebands_kmesh`.
#
# The QP energies are returned in r.qp_ebands_kpath and r.qp_ebands_kmesh.
# Note that the KS energies are optional but this is the recommended approach
# because the code will interpolate the corrections instead of the QP energies.

r = sigres.interpolate(lpratio=5,
                       ks_ebands_kpath=ks_ebands_kpath,
                       ks_ebands_kmesh=ks_ebands_kmesh
                       )
qp_edos = r.qp_ebands_kmesh.get_edos()

# Get points with ab-initio QP energies from the SIGRES so that
# we can plot the interpolate interpolated QP band structure with the first principles results.
# This part is optional
points = sigres.get_points_from_ebands(r.qp_ebands_kpath, size=24)
r.qp_ebands_kpath.plot(points=points, with_gaps=True)
#raise ValueError()

# Shortcut: pass the name of the GSR files directly.
#r = sigres.interpolate(ks_ebands_kpath=abidata.ref_file("si_nscf_GSR.nc"),
#                       ks_ebands_kmesh=abidata.ref_file("si_scf_GSR.nc"))
#ks_edos = r.ks_ebands_kmesh.get_edos()
#qp_edos = r.qp_ebands_kmesh.get_edos()

# Use ElectronBandsPlotter to plot the KS and the QP band structure with matplotlib.
plotter = ElectronBandsPlotter()
plotter.add_ebands("LDA", ks_ebands_kpath, edos=ks_edos)
plotter.add_ebands("GW (interpolated)", r.qp_ebands_kpath, edos=qp_edos)

# Get pandas dataframe with band structure parameters.
#df = plotter.get_ebands_frame()
#print(df)

# By default, the two band energies are shifted wrt to *their* fermi level.
# Use e=0 if you don't want to shift the eigenvalus
# so that it's possible to visualize the QP corrections.
plotter.combiplot(title="Combiplot")
plotter.boxplot(swarm=True, title="Boxplot")
plotter.combiboxplot(swarm=True, title="Combiboxplot")
# sphinx_gallery_thumbnail_number = 4
plotter.gridplot(title="Gridplot", with_gaps=True)

sigres.close()
"""
